# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.generators.python.typeset_generator.py
   :platform: Unix, Windows
   :synopsis: Generates code encapsulating an ontology's typeset.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.mp.ontologies import core
from pyesdoc.mp.ontologies.generators import generator_utils as gu
from pyesdoc.mp.ontologies.generators.python import utils as pgu
from pyesdoc.mp.ontologies.generators.generator import Generator
from pyesdoc.mp import utils



# Generator language.
_LANG = 'python'

# Template for a typeset module.
_TEMPLATE_TYPESET = 'typeset_for_package.txt'

# Template for a concrete class.
_TEMPLATE_CLASS_CONCRETE = "typeset_class_concrete.txt"

# Template for an abstract class.
_TEMPLATE_CLASS_ABSTRACT = "typeset_class_abstract.txt"

# Template for a class print string.
_TEMPLATE_CLASS_PSTR = "typeset_class_pstr.txt"

# Template for an enumeration.
_TEMPLATE_ENUM = "typeset_enum.txt"

# Template for a computed class property.
_TEMPLATE_CLASS_COMPUTED_PROPERTY = "typeset_class_computed_property.txt"

# Loaded templates.
_TEMPLATES = gu.load_templates(_LANG, (
    _TEMPLATE_TYPESET,
    _TEMPLATE_CLASS_CONCRETE,
    _TEMPLATE_CLASS_ABSTRACT,
    _TEMPLATE_CLASS_PSTR,
    _TEMPLATE_ENUM,
    _TEMPLATE_CLASS_COMPUTED_PROPERTY
))


class PackageTypeSetGenerator(Generator):
    """Generates code to represent an ontology as a set of types.

    """
    def on_package_parse(self, ctx):
        """Event handler for the package parse event.

        :param GeneratorContext ctx: Generation context information.

        """
        ctx.code.append(
            (
                _emit_module_typeset_for_pkg(ctx.ontology, ctx.pkg),
                pgu.get_ontology_directory(ctx),
                pgu.get_package_module_file_name(ctx.pkg, 'typeset')
            )
        )


def _emit_module_typeset_for_pkg(o, p):
    """Emits typeset module for an ontology package.

    """
    def emit_imports():
        def get_code(ap):
            return "import {} as {}{}".format(
                pgu.get_package_module_name(ap, 'typeset'),
                ap.op_name,
                gu.emit_line_return())

            return code

        return gu.emit(p.associated_for_import, get_code)


    def get_classes(p):
        # Return list of classes sorted in dependency order.
        result = [c for c in p.classes if c.base is None or c.base.package != p]
        result = sorted(result, key=lambda c: c.name)

        while len(result) < len(p.classes):
            n = len(result)
            for c in [c for c in sorted(p.classes, key=lambda c: c.name) if c not in result]:
                if c.base in result:
                    result.append(c)

            # Error if parse revels unresolvable dependencies.
            if n == len(result):
                msg = "Package {0} has circular class dependencies."
                msg = msg.format(p)
                utils.raise_error(msg.format(p))

        return result


    def emit_types():
        def get_code(t):
            if isinstance(t, core.Class):
                code = _emit_snippet_class(t)
            else:
                code = _emit_snippet_enum(t)
            code += gu.emit_line_return(2)

            return code

        return gu.emit(get_classes(p), get_code, sort=False) + \
               gu.emit(p.enums, get_code)


    code = _TEMPLATES[_TEMPLATE_TYPESET]
    code = code.replace('{imports}', emit_imports())
    code = code.replace('{types}', emit_types())
    code = code.replace('{package-name}', p.op_name)

    return code


def _emit_snippet_enum(e):
    """Emits code corresponding to a python enum.

    """
    def emit_members():
        def get_code(m):
            code = gu.emit_line_return()
            code += gu.emit_indent(2)
            code += '"{}"'.format(m.name)

            return code

        code = gu.emit(e.members, get_code, joiner=",", sort=False)
        code += gu.emit_line_return()
        code += gu.emit_indent(2)

        return code


    def emit_descriptions():
        def get_code(m):
            code = gu.emit_line_return()
            code += gu.emit_indent(2)
            code += '"{}"'.format(m.doc_string)

            return code

        code = gu.emit(e.members, get_code, joiner=",", sort=False)
        code += gu.emit_line_return()
        code += gu.emit_indent(2)

        return code


    code = _TEMPLATES[_TEMPLATE_ENUM]
    code = code.replace('{enum-name}', e.op_name)
    code = code.replace('{enum-doc-string}', e.doc_string)
    code = code.replace('{enum-is-open}', str(e.is_open))
    code = code.replace('{enum-members}', emit_members() if e.members else "")
    code = code.replace('{enum-descriptions}', emit_descriptions() if e.members else "")

    return code


def _emit_snippet_class(c):
    """Emits code corresponding to a python class."""
    # Open template.
    if c.is_abstract:
        code = _TEMPLATES[_TEMPLATE_CLASS_ABSTRACT]
    else:
        code = _TEMPLATES[_TEMPLATE_CLASS_CONCRETE]

    # Generate code.
    code = code.replace('{class-name}', c.op_name)
    code = code.replace('{base-class-name}', c.op_base_name)
    code = code.replace('{class-doc-string}', _emit_snippet_class_doc_string(c))
    code = code.replace('{class_constants}', _emit_snippet_class_property_constants(c))
    code = code.replace('{class-properties}', _emit_snippet_class_properties(c))
    code = code.replace('{class-computed-properties}', _emit_snippet_class_computed_properties(c))
    code = code.replace('{class-pstr}', _emit_snippet_class_pstr(c))

    return code


def _emit_snippet_class_pstr(c):
    """Emits class print string.

    """
    code = ""
    if c.pstr:
        code = _TEMPLATES[_TEMPLATE_CLASS_PSTR]
        code = code.replace('{text}', c.pstr.text)
        code = code.replace('{fields}', ", ".join(["self.{}".format(f) for f in c.pstr.fields]))

    return code


def _emit_snippet_class_doc_string(c):
    """Emits class doc string."""
    doc_string = "" if not c.doc_string else c.doc_string

    if doc_string:
        doc_string = "{0}{1}{2}".format(doc_string,
                                        gu.emit_line_return(2),
                                        gu.emit_indent())

    return "{0}{1}{2}".format(gu.emit_line_return(2),
                              gu.emit_indent(),
                              doc_string)



def _emit_snippet_class_properties(c):
    """Emits set of class properties."""
    def get_code(p):
        return "{0}{1}{2}# {3} ({4}){5}".format(
            gu.emit_indent(2),
            pgu.get_property_ctor(p),
            ''.ljust(50 - len(pgu.get_property_ctor(p))),
            pgu.get_type_doc_name(p.type),
            p.cardinality,
            gu.emit_line_return()
        )

    return gu.emit(c.properties, get_code)


def _emit_snippet_class_computed_properties(c):
    """Emits set of class computed properties."""
    def get_code(cp):
        # Open template.
        code = _TEMPLATES[_TEMPLATE_CLASS_COMPUTED_PROPERTY]

        # Generate code.
        code = code.replace('{computed-property-name}', cp.name)
        code = code.replace('{computed-property-computation}', cp.computation)

        return code

    return gu.emit(c.computed_properties, get_code)


def _emit_snippet_class_property_constants(c):
    """Emits set of class property constants."""
    def get_code(cnt):
        prp = c.get_property(cnt[0])
        if prp is not None:
            return '{0}self.{1} = {2}("{3}"){4}'.format(
                gu.emit_indent(2),
                cnt[0],
                pgu.get_type_functional_name(prp.type),
                cnt[1],
                gu.emit_line_return()
            )

    return gu.emit(c.constants, get_code)



# -*- coding: utf-8 -*-

"""
.. module:: test_serialization.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyesdoc document serialization tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
# Module imports.
import pyesdoc
import test_utils as tu
import test_types as tt



def _test(mod, doc, encoding):
    """Perform encoding specific serialization tests."""
    # Encode/decode and re-assert document.
    as_repr = tu.encode(doc, encoding)
    as_doc_1 = tu.decode(as_repr, encoding)
    tu.assert_doc(mod, as_doc_1)

    # Re-encode.
    as_repr_1 = tu.encode(as_doc_1, encoding)

    # TODO - explore why sometimes XML encoding is problematic
    # although all good when reencoding to json.
    if len(as_repr) != len(as_repr_1) and encoding == 'xml':
        as_repr = tu.encode(tu.decode(as_repr, encoding), 'json')
        as_repr_1 = tu.encode(tu.decode(as_repr_1, encoding), 'json')

    # Verify encoding equivalence.
    tu.assert_int(len(as_repr), len(as_repr_1), msg=str(type(doc)))
    tu.assert_str(as_repr, as_repr_1)


def _test_metafor_xml(doc):
    """Perform metafor cim v1 xml serialization tests."""
    # Simply verify that encoding is not unsupported at this time.
    try:
        tu.encode(doc, pyesdoc.METAFOR_CIM_XML_ENCODING)
    except NotImplementedError:
        pass


def _test_html(mod, doc):
    """Perform html serialization tests."""
    try:
        tu.encode(doc, pyesdoc.ENCODING_HTML)
    except KeyError as err:
        if mod.DOC_TYPE.type_key.lower() in (
            "cim.1.misc.documentset",
            "cim.1.software.statisticalmodelcomponent"):
            pass
        else:
            raise err


def test():
    """Performs serialization tests over the set of test documents."""
    for mod in tt.MODULES:
        # Get doc instance.
        doc = tu.get_doc(mod)

        # Test json.
        tu.init(_test, 'serialization', mod, 'JSON')
        yield _test, mod, doc, pyesdoc.ENCODING_JSON

        # Test xml.
        tu.init(_test, 'serialization', mod, 'XML')
        yield _test, mod, doc, pyesdoc.ENCODING_XML

        # Test metafor cim v1 xml.
        tu.init(_test_metafor_xml, 'serialization', mod, 'XML Metafor CIM v1')
        yield _test_metafor_xml, doc

        # Test html.
        tu.init(_test_html, 'serialization', mod, 'HTML')
        yield _test_html, mod, doc



# -*- coding: utf-8 -*-
"""
.. module:: downloader.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Downloads a document pulled from a 3rd party ATOM feed.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import datetime

import requests

import pyesdoc
from pyesdoc.utils import convert
from pyesdoc.utils import runtime



# Requests session - improves download times by pooling connection.
_SESSION = requests.Session()
_SESSION.mount('http://', requests.adapters.HTTPAdapter(
    pool_connections=100, pool_maxsize=100))


# Error file template.
_ERROR = """
------------------------------------------------------------
ES-DOC ARCHIVE ERROR @ {0}
------------------------------------------------------------
An error occurred whilst pulling a document from a remote source:

PROJECT = {1};
SOURCE = {2};
DOCUMENT URL = {3};
DOCUMENT ENCODING = {4};
ERROR TYPE = {5};
ERROR = {6}.

"""


class _DownloadException(Exception):
    """Exception raised when a download fails.

    """
    pass


class _EncodingException(Exception):
    """Exception raised when raw document encoding fails.

    """
    pass


class _IOException(Exception):
    """Exception raised when document io fails.

    """
    pass


class _ParsingException(Exception):
    """Exception raised when document parsing fails.

    """
    pass


def _log_start(ctx):
    """Writes a processing start message to standard output.

    """
    if ctx.verbose:
        msg = "processing doc {0}: {1} --> {2} --> {3}"
        pyesdoc.log(msg.format(ctx.index, ctx.project, ctx.source, ctx.entry.url))


def _download(ctx):
    """Attempts to download document from remote endpoint.

    """
    try:
        ctx.download = _SESSION.get(ctx.entry.url)
    except Exception as err:
        raise _DownloadException(err)
    else:
        if ctx.download.status_code != 200:
            raise _DownloadException()


def _set_content(ctx):
    """Set document content to be written to file system.

    """
    try:
        ctx.content = convert.unicode_to_str(ctx.download.text)
    except Exception as err:
        raise _EncodingException(err)


def _parse_content(ctx):
    """Parses document content.

    """
    try:
        ctx.content = pyesdoc.archive.parse(ctx.content,
                                            ctx.entry.encoding,
                                            ctx.project,
                                            ctx.source,
                                            "content")
    except Exception as err:
        raise _ParsingException(err)


def _write_content(ctx):
    """Writes downloaded document content to the file system.

    """
    try:
        with open(ctx.file.path, 'w') as output:
            output.write(ctx.content)
    except IOError as err:
        raise _IOException(err)


def _log_error(ctx):
    """Write processing error message to standard output.

    """
    msg = "ARCHIVE ERROR :: processing document {0}: {1} ---> {2}"
    msg = msg.format(ctx.index, ctx.project, ctx.error)
    pyesdoc.log_error(msg)


def _write_error(ctx):
    """Writes document processing error to file system.

    """
    try:
        with open(ctx.file.path_error, 'wb') as output:
            output.seek(0)
            output.write(_ERROR.format(
                datetime.datetime.now(),
                ctx.project,
                ctx.source,
                ctx.entry.url,
                ctx.entry.encoding,
                type(ctx.error),
                ctx.error
                ))
    except IOError:
        pyesdoc.log_error("Document processing error handling failed.")


def process(ctx):
    """Processes a document from a remote feed.

    :param object ctx: Document processing context.

    """
    tasks = (
        _log_start,
        _download,
        _set_content,
        _parse_content,
        _write_content
        )

    error_tasks = (
        _log_error,
        _write_error
        )

    runtime.invoke(ctx, tasks, error_tasks)



"""
.. module:: utils_constants.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Specialization model constants.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
# Specialization type keys.
TYPE_KEY_ENUM = 'enum'
TYPE_KEY_ENUM_CHOICE = 'enum-choice'
TYPE_KEY_GRID = "grid"
TYPE_KEY_KEYPROPS = "keyprops"
TYPE_KEY_MODEL = "model"
TYPE_KEY_PROCESS = "process"
TYPE_KEY_PROPERTY = "property"
TYPE_KEY_PROPERTY_SET = "property-set"
TYPE_KEY_REALM = "realm"
TYPE_KEY_SUBPROCESS = "subprocess"

# Set of abbreviations that do not require converting.
ABBREVIATIONS = {
	'CO2',
	'N20',
	'CH4'
}



# -*- coding: utf-8 -*-
"""
.. module:: pyesdoc.utils.__init__.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Package init.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.utils import convert
from pyesdoc.utils import runtime
from pyesdoc.utils import logger
from pyesdoc.utils import config



# -*- coding: utf-8 -*-

"""
.. module:: test_type_cim_v1_activity_ensemble_1.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Tests a cim.v1.Ensemble instance.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import arrow

import pyesdoc.ontologies.cim as cim
import test_utils as tu



# Test type.
DOC_TYPE = cim.v1.Ensemble

# Test type display name.
DOC_TYPE_DISPLAY_NAME = "Ensemble"

# Test document type.
DOC_TYPE_KEY = DOC_TYPE.type_key + '-1'

# Test representation file.
DOC_FILE = 'cim.1.activity.Ensemble-1.xml-metafor-cim-v1'

# Test document uid.
DOC_UID = '4c95e1c8-5d79-11e1-83af-00163e9152a5'

# Test document version.
DOC_VERSION = '1'

# Test document creation date.
DOC_DATE = arrow.get('2012-04-23 14:59:07.019144').datetime

# Test document project.
DOC_PROJECT = "CMIP5"

# Test document project.
DOC_INSTITUTE = "MOHC"

# Test document author.
DOC_AUTHOR = "Metafor Questionnaire"


def assert_extension_info(ext):
    """Asserts a document's extension information.

    :param object ext: Document extension information.

    """
    tu.assert_str(ext.display_name, u"amip")
    tu.assert_str(ext.description, u"Six element ensemble", True)
    tu.assert_str(ext.full_display_name, u"CMIP5 Ensemble : MOHC - amip")
    tu.assert_int(ext.summary_fields, 2)
    tu.assert_str(ext.summary_fields[0], u"amip")
    tu.assert_str(ext.summary_fields[1], u"3.3 AMIP and AMIP Ensemble")


def _assert_doc_core(doc, is_update):
    """Assert core information."""
    if not is_update:
        tu.assert_str(doc.description, "Six element ensemble", True)
    tu.assert_iter(doc.members, 6, cim.v1.EnsembleMember)
    tu.assert_str(doc.long_name, "3.3 AMIP and AMIP Ensemble")
    tu.assert_str(doc.short_name, "amip")
    tu.assert_iter(doc.types, 1)
    tu.assert_str(doc.types[0], "Initial Condition")

def _assert_doc_member(doc, is_update):
    """Assert ensemble members information."""
    # ... ensemble member id.
    member = doc.members[0]
    tu.assert_iter(member.ensemble_ids, 1, cim.v1.StandardName)
    tu.assert_str(member.ensemble_ids[0].is_open, True)
    tu.assert_str(member.ensemble_ids[0].value, "r1i1p1")

    # ... ensemble member simulation reference.
    tu.assert_object(member.simulation_reference, cim.v1.DocReference)
    ref = member.simulation_reference
    tu.assert_str(ref.description, "Reference to a simulation called amip")
    tu.assert_uuid(ref.id, "e4155a28-268e-11e1-9a0e-00163e9152a5")
    tu.assert_str(ref.name, "amip")
    tu.assert_str(ref.type, "simulation")
    tu.assert_str(ref.version, 2)


def assert_doc(doc, is_update=False):
    """Asserts a document.

    :param object doc: Document being tested.
    :param bool is_update: Flag indicating whether document has been updated.

    """
    for assertor in (
        _assert_doc_core,
        _assert_doc_member,
        ):
        assertor(doc, is_update)


def update_doc(doc):
    """Update a document prior to republishing.

    :param object doc: Document being republished.

    """
    doc.description = "X"


def assert_doc_updates(doc):
    """Asserts a document after being updated.

    :param object doc: Document being tested.

    """
    tu.assert_str(doc.description, "X")



"""
A realm key-properties sepecialization.
For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Ruth Petrie'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Ruth Petrie'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Sea Ice key properties'

# --------------------------------------------------------------------
# GENERAL details.
# --------------------------------------------------------------------
DETAILS['model'] = {
    'description': "Name of seaice model used.",
    'properties': [
        ('model_name', 'str', '1.1',
         "Name of seaice model (e.g. CICE 4.2, LIM 2.1, etc.)"),
    ]
}

DETAILS['variables'] = {
    'description': "List of prognostic variable in the sea ice model.",
    'properties': [
        ('prognostic', 'ENUM:prognostic_variables', '1.N',
            "List of prognostic variables in the sea ice component."),
        ]
    }

DETAILS['seawater_properties'] = {
    'description': "Properties of seawater relevant to sea ice",
    'properties': [
        ('ocean_freezing_point', 'ENUM:seawater_freezing_point', '1.1',
            "Equation used to compute the freezing point (in deg C) of seawater, as a function of salinity and pressure"),
        ('ocean_freezing_point_value', 'float', '0.1',
            "If using a constant seawater freezing point, specify this value."),
        ]
    }

# --------------------------------------------------------------------
# RESOLUTION: Description of the resolution in the sea ice grid.
# --------------------------------------------------------------------
DETAILS['resolution'] = {
    'description': "Resolution of the sea ice grid",
    'properties': [
        ('name', 'str', '1.1',
            "This is a string usually used by the modelling group to describe the resolution of this grid e.g. N512L180, T512L70, ORCA025 etc."),
        ('canonical_horizontal_resolution', 'str', '1.1',
            "Expression quoted for gross comparisons of resolution, eg. 50km or 0.1 degrees etc."),
        ('number_of_horizontal_gridpoints', 'int', '1.1',
            "Total number of horizontal (XY) points (or degrees of freedom) on computational grid."),
        ]
    }

# --------------------------------------------------------------------
# TUNING APPLIED: Any tuning used to optimise the parameters in this realm
# --------------------------------------------------------------------
DETAILS['tuning_applied'] = {
    'description': "Tuning applied to sea ice model component",
    'properties': [
        ('description', 'str', '1.1',
            "General overview description of tuning: explain and motivate the main targets and metrics retained.  Document the relative weight given to climate performance metrics versus process oriented metrics, and on the possible conflicts with parameterization level tuning. In particular describe any struggle with a parameter value that required pushing it to its limits to solve a particular model deficiency."),
        ('target', 'str', '1.1',
            "What was the aim of tuning, e.g. correct sea ice minima, correct seasonal cycle."),
        ('simulations', 'str', '1.1',
            "Which simulations had tuning applied, e.g. all, not historical, only pi-control? "),
        ('metrics_used', 'str', '1.1',
            "List any observed metrics used in tuning model/parameters"),
        ('variables', 'str', '0.1',
             "Which variables were changed during the tuning process?"),
        ]
    }

# --------------------------------------------------------------------
# ASSUMPTIONS: Any key assumptions made in this realm
# --------------------------------------------------------------------
DETAILS['assumptions'] = {
    'description': "Assumptions made in the sea ice model",
    'properties': [
        ('description', 'str', '1.N',
            "General overview description of any *key* assumptions made in this model."),
        ('on_diagnostic_variables', 'str', '1.N',
            "Note any assumptions that specifically affect the CMIP6 diagnostic sea ice variables."),
        ('missing_processes', 'str', '1.N',
             "List any *key* processes missing in this model configuration? Provide full details where this affects the CMIP6 diagnostic sea ice variables?"),
        ]
    }

# --------------------------------------------------------------------
# EXTRA CONSERVATION PROPERTIES: Details of methodology needed to conserve variables between processes
# --------------------------------------------------------------------
DETAILS['conservation'] = {
    'description': "Conservation in the sea ice component",
    'properties': [
        ('description', 'str', '1.1',
            "Provide a general description of conservation methodology."),
        ('properties', 'ENUM:conserved_properties', '1.N',
            "Properties conserved in sea ice by the numerical schemes."),
        ('budget', 'str', '1.1',
            "For each conserved property, specify the output variables which close the related budgets."),
        ('was_flux_correction_used', 'bool', '1.1',
            "Does conservation involved flux correction?"),
        ('corrected_conserved_prognostic_variables', 'str', '1.1',
            "List any variables which are conserved by *more* than the numerical scheme alone."),
        ]
    }

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['prognostic_variables'] = {
    'description': 'Prognostic variables in sea ice model',
    'is_open': True,
    'members': [
        ('Sea ice temperature', None),
        ('Sea ice concentration', None),
        ('Sea ice thickness', None),
        ('Sea ice volume per grid cell area', None),
        ('Sea ice u-velocity', None),
        ('Sea ice v-velocity', None),
        ('Sea ice enthalpy', None),
        ('Internal ice stress', None),
        ('Salinity', None),
        ('Snow temperature', 'Snow on ice temperature'),
        ('Snow depth', 'Snow on ice thickness'),
    ]
}

ENUMERATIONS['seawater_freezing_point'] = {
    'description': 'Types of seawater freezing point equation in sea ice.',
    'is_open': True,
    'members': [
        ('TEOS-10', 'Thermodynamic equation of seawater 2010'),
        ('Constant', 'Constant value of seawater freezing point is used.'),
        ]
}

ENUMERATIONS['conserved_properties'] = {
    'description': 'List of properties that can be conserved in sea ice',
    'is_open': True,
    'members': [
        ('Energy', None),
        ('Mass', None),
        ('Salt', None),
    ]
}



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.drq.content.cache.py
   :copyright: Copyright "February 27, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: An internal cache of decoded objects.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import collections


# Simple map of drq object uid's to drq object.
DATA = collections.OrderedDict()



def add(instance):
	"""Adds an instance to the cache.

	"""
	DATA[instance.uid] = instance



# -*- coding: utf-8 -*-

"""
.. module:: grids_classes.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Set of CIM v1 grids package class definitions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""



def coordinate_list():
    """The CoordList type may be used to specify a list of coordinates, typically for the purpose of defining coordinates along the X, Y or Z axes. The length of the coordinate list is given by the attribute of that name. This may be used by software to allocate memory in advance of storing the coordinate values. The hasConstantOffset attribute may be used to indicate that the coordinate list consists of values with constant offset (spacing). In this case only the first coordinate value and the offset (spacing) value need to be specified; however, the length attribute must still define the final as-built size of the coordinate list.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('has_constant_offset', 'bool', '0.1'),
            ('length', 'int', '0.1'),
            ('uom', 'str', '0.1'),
        ],
        'doc_strings': {
            'has_constant_offset': 'Set to true if coordinates in the built array have constant offset.',
            'length': 'Specifies the length of the coordinate array. This should always be the final, as-built length of the array if the hasConstantOffset property is set to true and the compact notation (start coordinate plus offset) is used.',
            'uom': 'Units of measure used by the coordinates.'
        },
        'decodings' : [
            ('has_constant_offset', '@hasConstantOffset'),
            ('length', '@length'),
            ('uom', '@uom'),
        ]
    }


def grid_extent():
    """DataType for recording the geographic extent of a gridMosaic or gridTile.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('minimum_latitude', 'str', '1.1'),
            ('maximum_latitude', 'str', '1.1'),
            ('minimum_longitude', 'str', '1.1'),
            ('maximum_longitude', 'str', '1.1'),
            ('units', 'str', '0.1'),
        ],
        'decodings' : [
            ('minimum_latitude', 'child::cim:latMin'),
            ('maximum_latitude', 'child::cim:latMax'),
            ('minimum_longitude', 'child::cim:lonMin'),
            ('maximum_longitude', 'child::cim:lonMax'),
            ('units', 'child::cim:units'),
        ]
    }


def grid_mosaic():
    """The GridMosaic class is used to define the geometry properties of an earth system model grid or an exchange grid. Such a grid definition may then be referenced by any number of earth system models. A GridMosaic object consists either of 1 or more child GridMosaics, or one or more child GridTiles, but not both. In the latter case the isLeaf property should be set to true, indicating that the mosaic is a leaf mosaic.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('citations', 'shared.citation', '0.N'),
            ('description', 'str', '0.1'),
            ('extent', 'grids.grid_extent', '0.1'),
            ('has_congruent_tiles', 'bool', '0.1'),
            ('id', 'str', '1.1'),
            ('is_leaf', 'bool', '1.1'),
            ('long_name', 'str', '0.1'),
            ('mnemonic', 'str', '0.1'),
            ('mosaic_count', 'int', '0.1'),
            ('mosaics', 'grids.grid_mosaic', '0.N'),
            ('short_name', 'str', '1.1'),
            ('tile_count', 'int', '0.1'),
            ('tiles', 'grids.grid_tile', '0.N'),
            ('type', 'str', '1.1'),
        ],
        'doc_strings': {
            'citations': 'Optional container element for specifying a list of references that describe the grid.',
            'description': 'A free-text description of a grid mosaic.',
            'has_congruent_tiles': 'Indicates whether or not all the tiles contained within a grid mosaic are congruent, that is, of the same size and shape.',
            'id': 'Specifies a globally unique identifier for a grid mosaic instance. By globally we mean across all GridSpec instances/records within a given modelling activity (such as CMIP5).',
            'is_leaf': 'Indicates whether or not a grid mosaic is a leaf mosaic, that is, it only contains child grid tiles not further mosaics.',
            'long_name': 'Specifies the long name associated with a grid mosaic. The long name will typically be a human-readable string, with acronyms expanded, used for labelling purposes.',
            'mosaic_count': 'Specifies the number of mosaics associated with a non-leaf grid mosaic. Set to zero if the grid mosaic is a leaf mosaic, i.e. it contains child grid tiles not mosaics.',
            'short_name': 'Specifies the short name associated with a grid mosaic. The short name will typically be a convenient abbreviation used to refer to a grid mosaic, e.g. \'UM ATM N96\'.',
            'tile_count': 'Specifies the number of tiles associated with a leaf grid mosaic. Set to zero if the grid mosaic is not a leaf mosaic, i.e. it contains child grid mosaics rather than tiles. (Added to align with equivalent ESG/Curator property.)',
            'type': 'Specifies the type of all the grid tiles contained in a grid mosaic. It is assumed that all of the tiles comprising a given grid mosaic are of the same type. The value domain is as per the specified enumeration list.'
        },
        'decodings' : [
            ('citations', 'child::cim:citationList/cim:citation'),
            ('description', 'child::cim:description'),
            ('has_congruent_tiles', '@congruentTiles'),
            ('id', '@id'),
            ('is_leaf', '@isLeaf'),
            ('long_name', 'child::cim:longName'),
            ('mnemonic', 'child::cim:mnemonic'),
            ('mosaic_count', '@numMosaics'),
            ('mosaics', 'child::cim:gridMosaic'),
            ('short_name', 'child::cim:shortName'),
            ('tile_count', '@numTiles'),
            ('tiles', 'child::cim:gridTile'),
            ('type', '@gridType'),
        ]
    }


def grid_property():
    """Creates and returns instance of grid_property class.

    """
    return {
        'type' : 'class',
        'base' : 'shared.property',
        'is_abstract' : False
    }


def grid_spec():
    """This is a container class for GridSpec objects. A GridSpec object can contain one or more esmModelGrid objects, and one or more esmExchangeGrid objects. These objects may be serialised to one or possibly several files according to taste. Since GridSpec is sub-typed from GML\'s AbstractGeometryType it can, and should, be identified using a gml:id attribute.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'is_document': True,
        'properties' : [
            ('esm_model_grids', 'grids.grid_mosaic', '0.N'),
            ('esm_exchange_grids', 'grids.grid_mosaic', '0.N'),
        ],
        'decodings' : [
            ('meta', 'self::cim:gridSpec'),
            ('esm_model_grids', 'child::cim:esmModelGrid'),
            ('esm_exchange_grids', 'child::cim:esmExchangeGrid'),
        ]
    }


def grid_tile():
    """The GridTile class is used to model an individual grid tile contained within a grid mosaic. A GridTile consists of an array of grid cells which may be defined in one of four ways: 1) for simple grids, by use of the SimpleGridGeometry data type; 2) by defining an array of GridCell objects; 3) by specifying an array of references to externally defined GridCell objects; or 4) by specifying a URI to a remote data file containing the grid cell definitions.  For all but the simplest grid tiles, it is envisaged that method 4 above will be the most frequently used option. However, it should be remembered that the CIM is primarily concerned with encoding climate model metadata. Specifying the coordinates of individual grid tiles and cells will most likely not be required as part of such metadata descriptions.  A GridTile object is associated with a geodetic or projected CRS via the horizontalCRS property, and with a vertical CRS via the verticalCRS property.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('area', 'str', '0.1'),
            ('cell_array', 'str', '0.1'),
            ('cell_ref_array', 'str', '0.1'),
            ('coord_file', 'str', '0.1'),
            ('coordinate_pole', 'str', '0.1'),
            ('horizontal_crs', 'str', '0.1'),
            ('grid_north_pole', 'str', '0.1'),
            ('simple_grid_geom', 'grids.simple_grid_geometry', '0.1'),
            ('vertical_crs', 'str', '0.1'),
            ('description', 'str', '0.1'),
            ('discretization_type', 'grids.discretization_enum', '0.1'),
            ('extent', 'grids.grid_extent', '0.1'),
            ('geometry_type', 'grids.geometry_type_enum', '0.1'),
            ('horizontal_resolution', 'grids.grid_tile_resolution_type', '0.1'),
            ('id', 'str', '0.1'),
            ('is_conformal', 'bool', '0.1'),
            ('is_regular', 'bool', '0.1'),
            ('is_terrain_following', 'bool', '0.1'),
            ('is_uniform', 'bool', '0.1'),
            ('long_name', 'str', '0.1'),
            ('mnemonic', 'str', '0.1'),
            ('nx', 'int', '0.1'),
            ('ny', 'int', '0.1'),
            ('nz', 'int', '0.1'),
            ('refinement_scheme', 'grids.refinement_type_enum', '0.1'),
            ('short_name', 'str', '0.1'),
            ('vertical_resolution', 'grids.grid_tile_resolution_type', '0.1'),
            ('zcoords', 'grids.vertical_coordinate_list', '0.1')
        ],
        'doc_strings' : {
            'area': 'gml:MeasureType:Specifies the area of the grid tile in the units defined by the uom attribute that is attached to the GML MeasureType data type.',
            'cell_array': 'GridCellArray:This property may be used to specify an array of grid cell definitions which together define the coordinate geometry of a grid tile. Depending on context, any of the existing sub-types of GridCell may be used. Mixing types is, however, not currently permitted.',
            'cell_ref_array': 'GridCellArray:This property may be used to define the coordinate geometry of a grid tile by specifying an array of references to remotely defined grid cells. Depending on context, any of the existing sub-types of GridCell may be referenced.',
            'coord_file': 'This property may be used to specify the URI of a file containing grid coordinates that define the geometry of a a grid tile. It is envisaged that this will be the preferred mechanism for specifying the geometry of complex grids.',
            'coordinate_pole': 'gml:PointType:The coordinatePole property may be used to specify the lat-long position of any coordinate poles (in the mathematical sense) that form part of the definition of a grid tile. Not to be confused with the gridNorthPole property.  If required, two or more coordinate pole definitions may be distinguished by setting the gml:id attribute to appropriate values, such as spole, npole, etc.',
            'horizontal_crs': 'gml:CRSPropertyType:Specifies the horizontal coordinate reference system used in the definition of the grid tile coordinates. This property should normally be an xlink reference to an external horizontal CRS definition (e.g. in a separate CRS dictionary). If required, however, the property may be defined in situ within a CIM document.',
            'grid_north_pole': 'gml:PointType:If required, defines the lat-long position of the north pole used by the grid tile in the case of rotated/displaced pole grids. Not to be confused with the coordinatePole property.',
            'simple_grid_geom': 'SimpleGridGeometry:This property may be used to define the coordinates of the nodes or cells making up a simple (i.e. uniform or regular) grid tile. More details are provided in the description of the SimpleGridGeometry data type.',
            'vertical_crs': 'gml:CRSPropertyType:Specifies the vertical coordinate reference system used in the definition of the grid tile coordinates. This property should normally be an xlink reference to an external vertical CRS definition (e.g. in a separate CRS dictionary). If required, however, the property may be defined in situ within a CIM document.',
            'description': 'A free-text description of a grid tile.',
            'discretization_type': 'Indicates the type of discretization applied to the grid tile, e.g. "logically_rectangular".',
            'geometry_type': 'Indicates the geometric figure used to approximate the figure of the Earth, e.g. "sphere".',
            'horizontal_resolution': 'Provides an indication of the approximate spatial sampling size of the grid tile, i.e. the size of the underlying grid cells. (Note: the maximum spatial resolution of the grid is twice the sampling size (e.g. 2 km for a 1 km x 1 km grid pitch).',
            'id': 'Specifies an identifer for a grid tile that is unique within its parent grid mosaic. It is not required for this identifier to be unique either across all mosaics in a GridSpec or across all GridSpecs, though if that were the case it would not be detrimental.',
            'is_conformal': 'This property is used to indicate if the grid tile is conformal, i.e. angle-preserving. If so, angles measured on the grid are equal to the equivalent angles on the Earth.',
            'is_regular': 'If true, indicates that the horizontal coordinates of the grid can be defined using 1D arrays (vectors). This means that grid node locations are defined by the cartesian product of the X/Lon and Y/Lat coordinate vectors. It also means that grid cells are logically rectangular (they may also be physically rectangular in the case of projected coordinates).',
            'is_terrain_following': 'Set to true if the vertical coordinate system is terrain-following even if, as is often the case, this only applies to the lower levels of the grid.',
            'is_uniform': 'If true, indicates that horizontal coordinates have fixed offsets in the X and Y directions. If the offset is the same in both directions then the grids are logically square, otherwise they are logically rectangular. The offsets can be specified by two scalar values (or three values in the case of 3D grids).',
            'nx': 'Specifies the length of the X, or longitude, dimension of the grid tile.',
            'ny': 'Specifies the length of the Y, or latitude, dimension of the grid tile.',
            'nz': 'Specifies the length of the Z, or height/level, dimension of the grid tile. The zcoords coordinate list property, if specified, should have this length.',
            'vertical_resolution': 'Provides an indication of the approximate resolution of the grid tile in the vertical dimension. (Added to align with corresponding ESG/Curator and DIF property).',
            'zcoords': 'This optional property may be used to specify the vertical coordinates (e.g. heights or model levels) at which a grid tile is utilised or realised. In the case of simple grid tiles the equivalent zcoords property on the SimpleGridGeometry data type would be used instead. The current property is intended to be used when the horizontal grid coordinates are defined by one of the other methods.'
        },
        'decodings' : [
            ('description', 'child::cim:description'),
            ('discretization_type', '@discretizationType'),
            ('extent', 'child::cim:extent'),
            ('geometry_type', '@geometryType'),
            ('horizontal_resolution', 'child::cim:horizontalResolution'),
            ('id', '@id'),
            ('is_conformal', '@isConformal'),
            ('is_regular', '@isRegular'),
            ('is_terrain_following', '@isTerrainFollowing'),
            ('is_uniform', '@isUniform'),
            ('long_name', 'child::cim:longName'),
            ('mnemonic', 'child::cim:mnemonic'),
            ('nx', '@nx'),
            ('ny', '@ny'),
            ('nz', '@nz'),
            ('refinement_scheme', '@refinementScheme'),
            ('short_name', 'child::cim:shortName'),
            ('vertical_resolution', 'child::cim:verticalResolution'),
            ('zcoords', 'child::cim:zcoords'),
        ]
    }


def grid_tile_resolution_type():
    """Provides a description and set of named properties for the horizontal or vertical resolution.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('description', 'str', '0.1'),
            ('properties', 'grids.grid_property', '0.N'),
        ],
        'doc_strings': {
            'description': 'A description of the resolution.'
        },
        'decodings' : [
            ('description', '@description'),
            ('properties', 'child::cim:property'),
        ]
    }

def simple_grid_geometry():
    """SimpleGridGeometry:This property may be used to define the coordinates of the nodes or cells making up a simple (i.e. uniform or regular) grid tile. More details are provided in the description of the SimpleGridGeometry data type.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('dim_order', 'str', '0.1'),
            ('is_mesh', 'bool', '0.1'),
            ('num_dims', 'int', '1.1'),
            ('xcoords', 'grids.coordinate_list', '1.1'),
            ('ycoords', 'grids.coordinate_list', '1.1'),
            ('zcoords', 'grids.coordinate_list', '0.1')
        ],
        'doc_strings': {
            'xcoords': 'X-Co-ordinates',
            'ycoords': 'Y-Co-ordinates',
            'zcoords': 'Z-Co-ordinates',
        }
    }


def vertical_coordinate_list():
    """There are some specific attributes that are associated with vertical coordinates.

    """
    return {
        'type' : 'class',
        'base' : 'grids.coordinate_list',
        'is_abstract' : False,
        'properties' : [
            ('form', 'str', '0.1'),
            ('properties', 'grids.grid_property', '0.N'),
            ('type', 'str', '0.1')
        ],
        'doc_strings': {
            'type': 'Specifies the length of the coordinate array. This should always be the final, as-built length of the array if the hasConstantOffset property is set to true and the compact notation (start coordinate plus offset) is used.',
            'form': 'Units of measure used by the coordinates.',
        },
        'decodings' : [
            ('properties', 'child::cim:property'),
            ('type', '@coordinateType'),
            ('form', '@coordinateForm'),
        ]
    }



# -*- coding: utf-8 -*-

from pyesdoc._codecs import dictionary
from pyesdoc._codecs import html
from pyesdoc._codecs import json
from pyesdoc._codecs import xml
from pyesdoc._codecs import xml_metafor_cim_v1




# -*- coding: utf-8 -*-

"""
.. module:: shared_classes.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Set of CIM v1 shared package class definitions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""



def change():
    """A description of [a set of] changes applied at a particular time, by a particular party, to a particular unit of metadata.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('author', 'shared.responsible_party', '0.1'),
            ('date', 'datetime', '0.1'),
            ('description', 'str', '0.1'),
            ('details', 'shared.change_property', '1.N'),
            ('name', 'str', '0.1'),
            ('type', 'shared.change_property_type', '0.1'),
        ],
        'doc_strings': {
            'author': 'The person that made the change.',
            'date': 'The date the change was implemented.',
            'name': 'A mnemonic for describing a particular change.',
        },
        'decodings': [
            ('author', 'child::cim:changeAuthor'),
            ('date', 'child::cim:changeDate'),
            ('description', 'child::cim:description'),
            ('details', 'child::cim:detail'),
            ('name', 'child::cim:name'),
            ('type', 'self::cim:change/@type'),
        ]
    }


def change_property():
    """A description of a single change applied to a single target.  Every ChangeProperty has a description, and may also have a name from a controlled vocabulary and a value.

    """
    return {
        'type': 'class',
        'base': 'shared.property',
        'is_abstract': False,
        'properties': [
            ('description', 'str', '0.1'),
            ('id', 'str', '0.1')
        ],
        'doc_strings': {
            'description': 'A text description of the change.  May be used in addition to, or instead of, the more formal description provided by the "value" attribute.',
        },
        'decodings': [
            ('description', 'child::cim:description'),
            ('id', 'child::cim:id'),
        ]
    }


def citation():
    """An academic reference to published work.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('alternative_title', 'str', '0.1'),
            ('collective_title', 'str', '0.1'),
            ('location', 'str', '0.1'),
            ('date', 'datetime', '0.1'),
            ('date_type', 'str', '0.1'),
            ('organisation', 'str', '0.1'),
            ('role', 'str', '0.1'),
            ('title', 'str', '0.1'),
            ('type', 'str', '0.1'),
        ],
        'decodings': [
            ('alternative_title', 'child::gmd:alternateTitle/gco:CharacterString'),
            ('alternative_title', 'child::gmd:alternateTitle'),
            ('collective_title', 'gmd:collectiveTitle/gco:CharacterString'),
            ('collective_title', 'gmd:collectiveTitle'),
            ('date', 'child::gmd:date/gmd:CI_Date/gmd:date/gco:Date'),
            ('date_type', 'child::gmd:date/gmd:CI_Date/gmd:dateType/gmd:CI_DateTypeCode/@codeListValue'),
            ('location', 'child::gmd:otherCitationDetails/gco:CharacterString'),
            ('location', 'child::gmd:otherCitationDetails'),
            ('title', 'child::gmd:title/gco:CharacterString'),
            ('title', 'child::gmd:title'),
            ('type', 'child::gmd:presentationForm/gmd:CI_PresentationFormCode/@codeListValue'),
        ]
    }


def compiler():
    """A description of a compiler used on a particular platform.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('environment_variables', 'str', '0.1'),
            ('language', 'str', '0.1'),
            ('name', 'str', '1.1'),
            ('options', 'str', '0.1'),
            ('type', 'shared.compiler_type', '0.1'),
            ('version', 'str', '1.1'),
        ],
        'doc_strings': {
            'environment_variables': 'The set of environment_variables used during compilation (recorded here as a single string rather than separate elements)',
            'options': 'The set of options used during compilation (recorded here as a single string rather than separate elements)',
        },
        'decodings': [
            ('environment_variables', 'child::cim:compilerEnvironmentVariables'),
            ('language', 'child::cim:compilerLanguage'),
            ('name', 'child::cim:compilerName'),
            ('options', 'child::cim:compilerOptions'),
            ('type', 'child::cim:compilerType'),
            ('version', 'child::cim:compilerVersion'),
        ]
    }


def data_source():
    """A DataSource can be realised by either a DataObject (file), a DataContent (variable), a Component (model), or a ComponentProperty (variable); all of those can supply data.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': True,
        'properties': [
            ('purpose', 'shared.data_purpose', '0.1'),
        ]
    }


def license():
    """A description of a license restricting access to a unit of data or software.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('name', 'str', '0.1'),
            ('contact', 'str', '0.1'),
            ('description', 'str', '0.1'),
            ('is_unrestricted', 'bool', '0.1'),
        ],
        'doc_strings': {
            'name': 'The name that the license goes by (ie: "GPL")',
            'contact': 'The point of contact for access to this artifact; may be either a person or an institution.',
            'description': 'A textual description of the license; might be the full text of the license, more likely to be a brief summary',
            'is_unrestricted': 'If unrestricted="true" then the artifact can be downloaded with no restrictions (ie: there are no administrative steps for the user to deal with; code or data can be downloaded and used directly).',
        }
    }


def machine():
    """A description of a machine used by a particular platform.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('cores_per_processor', 'int', '0.1'),
            ('description', 'str', '0.1'),
            ('interconnect', 'shared.interconnect_type', '0.1'),
            ('name', 'str', '1.1'),
            ('libraries', 'str', '0.N'),
            ('location', 'str', '0.1'),
            ('maximum_processors', 'int', '0.1'),
            ('operating_system', 'shared.operating_system_type', '0.1'),
            ('system', 'str', '0.1'),
            ('type', 'shared.machine_type', '0.1'),
            ('vendor', 'shared.machine_vendor_type', '0.1'),
            ('processor_type', 'shared.processor_type', '0.1'),
        ],
        'doc_strings': {
            'libraries': 'The libraries residing on this machine.',
        },
        'decodings': [
            ('cores_per_processor', 'child::cim:machineCoresPerProcessor'),
            ('description', 'child::cim:machineDescription'),
            ('interconnect', 'child::cim:machineInterconnect/@value'),
            ('name', 'child::cim:machineName'),
            ('libraries', 'child::cim:machineLibrary'),
            ('location', 'child::cim:machineLocation'),
            ('maximum_processors', 'child::cim:machineMaximumProcessors'),
            ('operating_system', 'child::cim:machineOperatingSystem/@value'),
            ('system', 'child::cim:machineSystem'),
            ('type', '@machineType'),
            ('vendor', 'child::cim:machineVendor/@value'),
            ('processor_type', 'child::cim:machineProcessorType/@value'),
        ]
    }


def machine_compiler_unit():
    """Associates a machine with a [set of] compilers.  This is a separate class in case a platform needs to specify more than one machine/compiler pair.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('compilers', 'shared.compiler', '0.N'),
            ('machine', 'shared.machine', '1.1'),
        ],
        'decodings': [
            ('compilers', 'child::cim:compiler'),
            ('machine', 'child::cim:machine'),
        ]
    }


def platform():
    """A platform is a description of resources used to deploy a component/simulation.  A platform pairs a machine with a (set of) compilers.  There is also a point of contact for the platform.

    """
    return {
        'type': 'class',
        'name': 'platform',
        'base': None,
        'is_abstract': False,
        'is_document': True,
        'properties': [
            ('contacts', 'shared.responsible_party', '0.N'),
            ('description', 'str', '0.1'),
            ('long_name', 'str', '0.1'),
            ('short_name', 'str', '1.1'),
            ('units', 'shared.machine_compiler_unit', '1.N'),
        ],
        'decodings': [
            ('meta', 'self::cim:platform'),
            ('contacts', 'child::cim:contact'),
            ('description', 'child::cim:description'),
            ('long_name', 'child::cim:longName'),
            ('short_name', 'child::cim:shortName'),
            ('units', 'child::cim:unit'),
        ]
    }


def property():
    """A simple name/value pair representing a property of some entity or other.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('name', 'str', '0.1'),
            ('value', 'str', '0.1'),
        ],
        'decodings': [
            ('name', 'child::cim:name'),
            ('value', 'child::cim:value'),
        ]
    }


def relationship():
    """A record of a relationship between one document and another. This class is abstract; specific document types must specialise this class for their relationshipTypes to be included in a document\'s genealogy.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': True,
        'properties': [
            ('description', 'str', '0.1')
        ]
    }


def responsible_party():
    """A person/organsiation responsible for some aspect of a climate science artefact.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('abbreviation', 'str', '0.1'),
            ('address', 'str', '0.1'),
            ('email', 'str', '0.1'),
            ('individual_name', 'str', '0.1'),
            ('organisation_name', 'str', '0.1'),
            ('role', 'str', '0.1'),
            ('url', 'str', '0.1'),
        ],
        'decodings': [
            ('abbreviation', 'child::cim:abbreviation'),
            ('address', 'child::gmd:contactInfo/gmd:CI_Contact/gmd:address/gmd:CI_Address/gmd:deliveryPoint/gco:CharacterString'),
            ('address', 'child::gmd:contactInfo/gmd:CI_Contact/gmd:address/gmd:CI_Address/gmd:deliveryPoint'),
            ('email', 'child::gmd:contactInfo/gmd:CI_Contact/gmd:address/gmd:CI_Address/gmd:electronicMailAddress/gco:CharacterString'),
            ('email', 'child::gmd:contactInfo/gmd:CI_Contact/gmd:address/gmd:CI_Address/gmd:electronicMailAddress'),
            ('individual_name', 'child::gmd:individualName/gco:CharacterString'),
            ('individual_name', 'child::gmd:individualName'),
            ('organisation_name', 'child::gmd:organisationName/gco:CharacterString'),
            ('organisation_name', 'child::gmd:organisationName'),
            ('role', 'gmd:role/gmd:CI_RoleCode/@codeListValue'),
            ('url', 'child::gmd:contactInfo/gmd:CI_Contact/gmd:onlineResource/gmd:CI_OnlineResource/gmd:linkage/gmd:URL'),
        ]
    }


def standard():
    """Describes a name given to an entity from a recognised standard.  The CIM records the standard and the name.  For example, the standard might be CF and the name might be atmospheric_pressure.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('name', 'str', '1.1'),
            ('version', 'str', '0.1'),
            ('description', 'str', '0.1'),
        ],
        'doc_strings': {
            'name': 'The name of the standard',
            'version': 'The version of the standard',
            'description': 'The version of the standard'
        },
        'decodings': [
            ('name', 'child::cim:name'),
            ('version', 'child::cim:version'),
            ('description', 'child::cim:description'),
        ]
    }


def standard_name():
    """Describes a name given to an entity from a recognised standard.  The CIM records the standard and the name.  For example, the standard might be CF and the name might be atmospheric_pressure.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('is_open', 'bool', '1.1'),
            ('value', 'str', '1.1'),
            ('standards', 'shared.standard', '0.N'),
        ],
        'doc_strings': {
            'standards': 'Details of the standard being used.'
        },
        'decodings': [
            ('is_open', '@open'),
            ('value', '@value'),
            ('standards', 'child::cim:standard'),
        ]
    }





# -*- coding: utf-8 -*-

"""
.. module:: misc_classes.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Set of CIM v1 miscellaneous package class definitions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""



def document_set():
    """Represents a bundled set of documents.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'is_document': True,
        'properties' : [
            ('data', 'data.data_object', '0.N'),
            ('ensembles', 'activity.ensemble', '0.N'),
            ('experiment', 'activity.numerical_experiment', '0.1'),
            ('grids', 'grids.grid_spec', '0.N'),
            ('model', 'software.model_component', '0.1'),
            ('platform', 'shared.platform', '0.1'),
            ('simulation', 'activity.simulation_run', '0.1')
        ],
        'doc_strings': {
            'data': 'Associated input/output data.',
            'ensembles': 'Associated ensemble runs.',
            'experiment': 'Associated numerical experiment.',
            'grids': 'Associated grid-spec.',
            'model': 'Associated model component.',
            'platform': 'Associated simulation execution platform.',
            'simulation': 'Associated simulation run.'
        },
        'decodings' : [
            ('data', 'child::cim:dataObject'),
            ('ensembles', 'child::cim:ensemble'),
            ('experiment', 'child::cim:numericalExperiment'),
            ('model', 'child::cim:modelComponent'),
            ('grids', 'child::cim:gridSpec'),
            ('platform', 'child::cim:platform'),
            ('simulation', 'child::cim:simulationRun'),
        ]
    }



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.drq.content.section.py
   :copyright: Copyright "February 27, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Wraps a section defined within dreq.xml.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
class Section(object):
    """Wraps a section defined within dreq.xml.

    """
    def __init__(self, table):
        """Instance constructor.

        :param dreq.definition.Table table: Associated definition info.

        """
        self._TABLE = table
        self._items = []


    def __repr__(self):
        """Instance representation.

        """
        return repr(self._TABLE)


    def __len__(self):
        """Returns number of items in managed collection.

        """
        return len(self._items)


    def __iter__(self):
        """Instance iterator initializer.

        """
        return iter(sorted(self._items, key=lambda i: i._sort_key))


    def __getattr__(self, key):
        """Returns a child section item.

        """
        return self._get_item(key)


    def __getitem__(self, key):
        """Returns a child section item.

        """
        return self._get_item(key)


    def _get_item(self, key):
        """Returns a wrapped keyed value.

        """
        if isinstance(key, (slice, int)):
            return self._items[key]
        key = unicode(key).strip().lower()
        for item in self._items:
            if item.label.lower() == key:
                return item



# -*- coding: utf-8 -*-

"""
.. module:: cim.v2.typeset_for_data_package.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: The set of types of the cim.v2.data package.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
import abc
import datetime
import uuid

import typeset_for_activity_package as activity
import typeset_for_shared_package as shared



class Dataset(object):
    """A concrete class within the cim v2 type system.

    Dataset discovery information.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Dataset, self).__init__()

        self.availability = []                            # shared.OnlineResource (0.N)
        self.citations = []                               # shared.Citation (0.N)
        self.description = None                           # unicode (0.1)
        self.drs_datasets = []                            # drs.DrsPublicationDataset (0.N)
        self.meta = shared.DocMetaInfo()                  # shared.DocMetaInfo (1.1)
        self.name = None                                  # unicode (1.1)
        self.produced_by = None                           # data.Simulation (0.1)
        self.related_to_dataset = []                      # shared.OnlineResource (0.N)
        self.responsible_parties = []                     # shared.Responsibility (0.N)


    @property
    def __str__(self):
	    """Instrance string representation.

	    """
	    return "{}".format(self.name)


class InputDataset(object):
    """A concrete class within the cim v2 type system.

    An input dataset is used as within another component (such as a
model). It comprises an original, source dataset plus any
modifications requirted to use it in the relevant component.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(InputDataset, self).__init__()

        self.modifications_applied = None                 # unicode (0.1)
        self.original_data = None                         # data.Dataset (1.1)


class Simulation(activity.Activity):
    """A concrete class within the cim v2 type system.

    Simulation class provides the integrating link about what models
    were run and wny.  In many cases this should be auto-generated
    from output file headers.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Simulation, self).__init__()

        self.calendar = None                              # time.Calendar (0.1)
        self.end_time = None                              # time.DateTime (0.1)
        self.extra_attributes = []                        # shared.ExtraAttribute (0.N)
        self.forcing_index = None                         # int (0.1)
        self.further_info_url = None                      # unicode (0.1)
        self.initialization_index = None                  # int (0.1)
        self.insitution = None                            # shared.Party (0.1)
        self.parent_simulation = None                     # activity.ParentSimulation (0.1)
        self.part_of_project = []                         # designing.Project (1.N)
        self.physics_index = None                         # int (0.1)
        self.primary_ensemble = None                      # activity.Ensemble (0.1)
        self.ran_for_experiments = []                     # designing.NumericalExperiment (1.N)
        self.realization_index = None                     # int (0.1)
        self.start_time = None                            # time.DateTime (0.1)
        self.sub_experiment = None                        # designing.NumericalExperiment (0.1)
        self.used = None                                  # science.Model (1.1)
        self.variant_info = None                          # unicode (0.1)


class VariableCollection(object):
    """A concrete class within the cim v2 type system.

    A collection of variables within the scope of a code or process element.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(VariableCollection, self).__init__()

        self.collection_name = None                       # unicode (0.1)
        self.variables = []                               # unicode (1.N)


class Downscaling(Simulation):
    """A concrete class within the cim v2 type system.

    Defines a downscaling activity.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Downscaling, self).__init__()

        self.downscaled_from = None                       # data.Simulation (1.1)


class DataAssociationTypes(object):
    """An enumeration within the cim v2 type system.

    Set of possible dataset associations.

    Selected from, and extended from,  ISO19115 (2014) DS_AssociationTypeCode.
    """
    is_open = False
    members = [
        "revisonOf",
        "partOf",
        "isComposedOf"
        ]
    descriptions = [
        "This dataset was revised from the target",
        "This dataset forms part of the target",
        "This dataset is composed from the target"
        ]





# -*- coding: utf-8 -*-

"""
.. module:: cim.v1.decoder_for_activity_package.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: A set of cim 1 decoders.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
from decoder_xml_utils import set_attributes
from decoder_for_data_package import *
from decoder_for_shared_package import *
from decoder_for_software_package import *
import typeset



def decode_activity(xml, nsmap):
    """Decodes an instance of the following type: activity.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.Activity

    """
    decodings = [
        ('funding_sources', True, 'unicode', 'child::cim:fundingSource'),
        ('projects', True, 'unicode', 'child::cim:project/@value'),
        ('rationales', True, 'unicode', 'child::cim:rationale'),
        ('responsible_parties', True, decode_responsible_party, 'child::cim:responsibleParty'),
    ]

    return set_attributes(typeset.activity.Activity(), xml, nsmap, decodings)


def decode_boundary_condition(xml, nsmap):
    """Decodes an instance of the following type: boundary condition.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.BoundaryCondition

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('id', False, 'unicode', 'child::cim:id'),
        ('name', False, 'unicode', 'child::cim:name'),
        ('options', True, decode_numerical_requirement_option, 'child::cim:requirementOption'),
        ('source', False, decode_component_property, 'child::cim:source/cim:source/cim:componentProperty'),
        ('source', False, decode_data_content, 'child::cim:source/cim:source/cim:dataContent'),
        ('source', False, decode_data_object, 'child::cim:source/cim:source/cim:dataObject'),
        ('source', False, decode_doc_reference, 'child::cim:source/cim:reference'),
        ('source', False, decode_model_component, 'child::cim:source/cim:source/cim:softwareComponent'),
        ('source', False, decode_processor_component, 'child::cim:source/cim:source/cim:softwareComponent'),
        ('source', False, decode_statistical_model_component, 'child::cim:source/cim:source/cim:softwareComponent'),
    ]

    return set_attributes(typeset.activity.BoundaryCondition(), xml, nsmap, decodings)


def decode_conformance(xml, nsmap):
    """Decodes an instance of the following type: conformance.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.Conformance

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('frequency', False, 'unicode', 'child::cim:frequency'),
        ('is_conformant', False, 'bool', '@conformant'),
        ('requirements', True, decode_boundary_condition, 'child::cim:requirement/cim:requirement/cim:boundaryCondition'),
        ('requirements', True, decode_doc_reference, 'child::cim:requirement/cim:reference'),
        ('requirements', True, decode_initial_condition, 'child::cim:requirement/cim:requirement/cim:initialCondition'),
        ('requirements', True, decode_lateral_boundary_condition, 'child::cim:requirement/cim:requirement/cim:lateralBoundaryCondition'),
        ('requirements', True, decode_output_requirement, 'child::cim:requirement/cim:requirement/cim:outputRequirement'),
        ('requirements', True, decode_spatio_temporal_constraint, 'child::cim:requirement/cim:requirement/cim:spatioTemporalConstraint'),
        ('sources', True, decode_component_property, 'child::cim:source/cim:source/cim:componentProperty'),
        ('sources', True, decode_data_content, 'child::cim:source/cim:source/cim:dataContent'),
        ('sources', True, decode_data_object, 'child::cim:source/cim:source/cim:dataObject'),
        ('sources', True, decode_doc_reference, 'child::cim:source/cim:reference'),
        ('sources', True, decode_model_component, 'child::cim:source/cim:source/cim:softwareComponent'),
        ('sources', True, decode_processor_component, 'child::cim:source/cim:source/cim:softwareComponent'),
        ('sources', True, decode_statistical_model_component, 'child::cim:source/cim:source/cim:softwareComponent'),
        ('type', False, 'unicode', '@type'),
    ]

    return set_attributes(typeset.activity.Conformance(), xml, nsmap, decodings)


def decode_downscaling_simulation(xml, nsmap):
    """Decodes an instance of the following type: downscaling simulation.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.DownscalingSimulation

    """
    decodings = [
        ('calendar', False, decode_daily_360, 'child::cim:calendar/cim:daily-360'),
        ('calendar', False, decode_perpetual_period, 'child::cim:calendar/cim:perpetualPeriod'),
        ('calendar', False, decode_real_calendar, 'child::cim:calendar/cim:realCalendar'),
        ('description', False, 'unicode', 'child::cim:description'),
        ('downscaled_from', False, decode_component_property, 'child::cim:downscaledFrom/cim:downscaledFrom/cim:componentProperty'),
        ('downscaled_from', False, decode_data_content, 'child::cim:downscaledFrom/cim:downscaledFrom/cim:dataContent'),
        ('downscaled_from', False, decode_data_object, 'child::cim:downscaledFrom/cim:downscaledFrom/cim:dataObject'),
        ('downscaled_from', False, decode_doc_reference, 'child::cim:downscaledFrom/cim:reference'),
        ('downscaled_from', False, decode_model_component, 'child::cim:downscaledFrom/cim:downscaledFrom/cim:softwareComponent'),
        ('downscaled_from', False, decode_processor_component, 'child::cim:downscaledFrom/cim:downscaledFrom/cim:softwareComponent'),
        ('downscaled_from', False, decode_statistical_model_component, 'child::cim:downscaledFrom/cim:downscaledFrom/cim:softwareComponent'),
        ('downscaling_id', False, 'unicode', 'child::cim:downscalingID'),
        ('downscaling_type', False, 'unicode', 'self::cim:downscalingSimulation/@downscalingType'),
        ('funding_sources', True, 'unicode', 'child::cim:fundingSource'),
        ('inputs', True, decode_coupling, 'child::cim:input'),
        ('long_name', False, 'unicode', 'child::cim:longName'),
        ('meta', False, decode_doc_meta_info, 'self::cim:downscalingSimulation'),
        ('outputs', True, decode_data_object, 'child::cim:output/cim:dataObject'),
        ('outputs', True, decode_doc_reference, 'child::cim:output/cim:reference'),
        ('projects', True, 'unicode', 'child::cim:project/@value'),
        ('rationales', True, 'unicode', 'child::cim:rationale'),
        ('responsible_parties', True, decode_responsible_party, 'child::cim:responsibleParty'),
        ('short_name', False, 'unicode', 'child::cim:shortName'),
        ('supports', True, decode_doc_reference, 'child::cim:supports/cim:reference'),
        ('supports', True, decode_experiment, 'child::cim:supports/cim:experiment'),
    ]

    return set_attributes(typeset.activity.DownscalingSimulation(), xml, nsmap, decodings)


def decode_ensemble(xml, nsmap):
    """Decodes an instance of the following type: ensemble.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.Ensemble

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('funding_sources', True, 'unicode', 'child::cim:fundingSource'),
        ('long_name', False, 'unicode', 'child::cim:longName'),
        ('members', True, decode_ensemble_member, 'child::cim:ensembleMember'),
        ('meta', False, decode_doc_meta_info, 'self::cim:ensemble'),
        ('outputs', True, decode_component_property, 'child::cim:output/cim:output/cim:componentProperty'),
        ('outputs', True, decode_data_content, 'child::cim:output/cim:output/cim:dataContent'),
        ('outputs', True, decode_data_object, 'child::cim:output/cim:output/cim:dataObject'),
        ('outputs', True, decode_doc_reference, 'child::cim:output/cim:reference'),
        ('outputs', True, decode_model_component, 'child::cim:output/cim:output/cim:softwareComponent'),
        ('outputs', True, decode_processor_component, 'child::cim:output/cim:output/cim:softwareComponent'),
        ('outputs', True, decode_statistical_model_component, 'child::cim:output/cim:output/cim:softwareComponent'),
        ('projects', True, 'unicode', 'child::cim:project/@value'),
        ('rationales', True, 'unicode', 'child::cim:rationale'),
        ('responsible_parties', True, decode_responsible_party, 'child::cim:responsibleParty'),
        ('short_name', False, 'unicode', 'child::cim:shortName'),
        ('supports', True, decode_doc_reference, 'child::cim:supports/cim:reference'),
        ('supports', True, decode_experiment, 'child::cim:supports/cim:experiment'),
        ('types', True, 'unicode', 'child::cim:ensembleType/@value'),
    ]

    return set_attributes(typeset.activity.Ensemble(), xml, nsmap, decodings)


def decode_ensemble_member(xml, nsmap):
    """Decodes an instance of the following type: ensemble member.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.EnsembleMember

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('ensemble', False, decode_doc_reference, 'child::cim:ensemble/cim:reference'),
        ('ensemble', False, decode_ensemble, 'child::cim:ensemble/cim:ensemble'),
        ('ensemble_ids', True, decode_standard_name, 'child::cim:ensembleMemberID'),
        ('funding_sources', True, 'unicode', 'child::cim:fundingSource'),
        ('long_name', False, 'unicode', 'child::cim:longName'),
        ('projects', True, 'unicode', 'child::cim:project/@value'),
        ('rationales', True, 'unicode', 'child::cim:rationale'),
        ('responsible_parties', True, decode_responsible_party, 'child::cim:responsibleParty'),
        ('short_name', False, 'unicode', 'child::cim:shortName'),
        ('simulation', False, decode_doc_reference, 'child::cim:simulation/cim:reference'),
        ('simulation', False, decode_simulation, 'child::cim:ensemble/cim:simulation'),
        ('supports', True, decode_doc_reference, 'child::cim:supports/cim:reference'),
        ('supports', True, decode_experiment, 'child::cim:supports/cim:experiment'),
    ]

    return set_attributes(typeset.activity.EnsembleMember(), xml, nsmap, decodings)


def decode_experiment(xml, nsmap):
    """Decodes an instance of the following type: experiment.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.Experiment

    """
    decodings = [
        ('funding_sources', True, 'unicode', 'child::cim:fundingSource'),
        ('projects', True, 'unicode', 'child::cim:project/@value'),
        ('rationales', True, 'unicode', 'child::cim:rationale'),
        ('responsible_parties', True, decode_responsible_party, 'child::cim:responsibleParty'),
    ]

    return set_attributes(typeset.activity.Experiment(), xml, nsmap, decodings)


def decode_experiment_relationship(xml, nsmap):
    """Decodes an instance of the following type: experiment relationship.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.ExperimentRelationship

    """
    decodings = [
    ]

    return set_attributes(typeset.activity.ExperimentRelationship(), xml, nsmap, decodings)


def decode_experiment_relationship_target(xml, nsmap):
    """Decodes an instance of the following type: experiment relationship target.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.ExperimentRelationshipTarget

    """
    decodings = [
    ]

    return set_attributes(typeset.activity.ExperimentRelationshipTarget(), xml, nsmap, decodings)


def decode_initial_condition(xml, nsmap):
    """Decodes an instance of the following type: initial condition.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.InitialCondition

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('id', False, 'unicode', 'child::cim:id'),
        ('name', False, 'unicode', 'child::cim:name'),
        ('options', True, decode_numerical_requirement_option, 'child::cim:requirementOption'),
        ('source', False, decode_component_property, 'child::cim:source/cim:source/cim:componentProperty'),
        ('source', False, decode_data_content, 'child::cim:source/cim:source/cim:dataContent'),
        ('source', False, decode_data_object, 'child::cim:source/cim:source/cim:dataObject'),
        ('source', False, decode_doc_reference, 'child::cim:source/cim:reference'),
        ('source', False, decode_model_component, 'child::cim:source/cim:source/cim:softwareComponent'),
        ('source', False, decode_processor_component, 'child::cim:source/cim:source/cim:softwareComponent'),
        ('source', False, decode_statistical_model_component, 'child::cim:source/cim:source/cim:softwareComponent'),
    ]

    return set_attributes(typeset.activity.InitialCondition(), xml, nsmap, decodings)


def decode_lateral_boundary_condition(xml, nsmap):
    """Decodes an instance of the following type: lateral boundary condition.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.LateralBoundaryCondition

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('id', False, 'unicode', 'child::cim:id'),
        ('name', False, 'unicode', 'child::cim:name'),
        ('options', True, decode_numerical_requirement_option, 'child::cim:requirementOption'),
        ('source', False, decode_component_property, 'child::cim:source/cim:source/cim:componentProperty'),
        ('source', False, decode_data_content, 'child::cim:source/cim:source/cim:dataContent'),
        ('source', False, decode_data_object, 'child::cim:source/cim:source/cim:dataObject'),
        ('source', False, decode_doc_reference, 'child::cim:source/cim:reference'),
        ('source', False, decode_model_component, 'child::cim:source/cim:source/cim:softwareComponent'),
        ('source', False, decode_processor_component, 'child::cim:source/cim:source/cim:softwareComponent'),
        ('source', False, decode_statistical_model_component, 'child::cim:source/cim:source/cim:softwareComponent'),
    ]

    return set_attributes(typeset.activity.LateralBoundaryCondition(), xml, nsmap, decodings)


def decode_measurement_campaign(xml, nsmap):
    """Decodes an instance of the following type: measurement campaign.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.MeasurementCampaign

    """
    decodings = [
        ('funding_sources', True, 'unicode', 'child::cim:fundingSource'),
        ('projects', True, 'unicode', 'child::cim:project/@value'),
        ('rationales', True, 'unicode', 'child::cim:rationale'),
        ('responsible_parties', True, decode_responsible_party, 'child::cim:responsibleParty'),
    ]

    return set_attributes(typeset.activity.MeasurementCampaign(), xml, nsmap, decodings)


def decode_numerical_activity(xml, nsmap):
    """Decodes an instance of the following type: numerical activity.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.NumericalActivity

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('funding_sources', True, 'unicode', 'child::cim:fundingSource'),
        ('long_name', False, 'unicode', 'child::cim:longName'),
        ('projects', True, 'unicode', 'child::cim:project/@value'),
        ('rationales', True, 'unicode', 'child::cim:rationale'),
        ('responsible_parties', True, decode_responsible_party, 'child::cim:responsibleParty'),
        ('short_name', False, 'unicode', 'child::cim:shortName'),
        ('supports', True, decode_doc_reference, 'child::cim:supports/cim:reference'),
        ('supports', True, decode_experiment, 'child::cim:supports/cim:experiment'),
    ]

    return set_attributes(typeset.activity.NumericalActivity(), xml, nsmap, decodings)


def decode_numerical_experiment(xml, nsmap):
    """Decodes an instance of the following type: numerical experiment.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.NumericalExperiment

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('experiment_id', False, 'unicode', 'child::cim:experimentID'),
        ('funding_sources', True, 'unicode', 'child::cim:fundingSource'),
        ('long_name', False, 'unicode', 'child::cim:longName'),
        ('meta', False, decode_doc_meta_info, 'self::cim:numericalExperiment'),
        ('projects', True, 'unicode', 'child::cim:project/@value'),
        ('rationales', True, 'unicode', 'child::cim:rationale'),
        ('requirements', True, decode_boundary_condition, 'child::cim:numericalRequirement/cim:boundaryCondition'),
        ('requirements', True, decode_boundary_condition, 'child::cim:numericalRequirement[@xsi:type="BoundaryCondition"]'),
        ('requirements', True, decode_initial_condition, 'child::cim:numericalRequirement/cim:initialCondition'),
        ('requirements', True, decode_initial_condition, 'child::cim:numericalRequirement[@xsi:type="InitialCondition"]'),
        ('requirements', True, decode_lateral_boundary_condition, 'child::cim:numericalRequirement/cim:lateralBoundaryCondition'),
        ('requirements', True, decode_lateral_boundary_condition, 'child::cim:numericalRequirement[@xsi:type="LateralBoundaryCondition"]'),
        ('requirements', True, decode_output_requirement, 'child::cim:numericalRequirement/cim:outputRequirement'),
        ('requirements', True, decode_output_requirement, 'child::cim:numericalRequirement[@xsi:type="OutputRequirement"]'),
        ('requirements', True, decode_spatio_temporal_constraint, 'child::cim:numericalRequirement/cim:spatioTemporalConstraint'),
        ('requirements', True, decode_spatio_temporal_constraint, 'child::cim:numericalRequirement[@xsi:type="SpatioTemporalConstraint"]'),
        ('responsible_parties', True, decode_responsible_party, 'child::cim:responsibleParty'),
        ('short_name', False, 'unicode', 'child::cim:shortName'),
    ]

    return set_attributes(typeset.activity.NumericalExperiment(), xml, nsmap, decodings)


def decode_numerical_requirement(xml, nsmap):
    """Decodes an instance of the following type: numerical requirement.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.NumericalRequirement

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('id', False, 'unicode', 'child::cim:id'),
        ('name', False, 'unicode', 'child::cim:name'),
        ('options', True, decode_numerical_requirement_option, 'child::cim:requirementOption'),
        ('source', False, decode_component_property, 'child::cim:source/cim:source/cim:componentProperty'),
        ('source', False, decode_data_content, 'child::cim:source/cim:source/cim:dataContent'),
        ('source', False, decode_data_object, 'child::cim:source/cim:source/cim:dataObject'),
        ('source', False, decode_doc_reference, 'child::cim:source/cim:reference'),
        ('source', False, decode_model_component, 'child::cim:source/cim:source/cim:softwareComponent'),
        ('source', False, decode_processor_component, 'child::cim:source/cim:source/cim:softwareComponent'),
        ('source', False, decode_statistical_model_component, 'child::cim:source/cim:source/cim:softwareComponent'),
    ]

    return set_attributes(typeset.activity.NumericalRequirement(), xml, nsmap, decodings)


def decode_numerical_requirement_option(xml, nsmap):
    """Decodes an instance of the following type: numerical requirement option.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.NumericalRequirementOption

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('description', False, 'unicode', 'child::cim:requirement/cim:requirement/cim:boundaryCondition/cim:description'),
        ('description', False, 'unicode', 'child::cim:requirement/cim:requirement/cim:initialCondition/cim:description'),
        ('description', False, 'unicode', 'child::cim:requirement/cim:requirement/cim:lateralBoundaryCondition/cim:description'),
        ('description', False, 'unicode', 'child::cim:requirement/cim:requirement/cim:outputRequirement/cim:description'),
        ('description', False, 'unicode', 'child::cim:requirement/cim:requirement/cim:spatioTemporalConstraint/cim:description'),
        ('id', False, 'unicode', 'child::cim:id'),
        ('id', False, 'unicode', 'child::cim:requirement/cim:requirement/cim:boundaryCondition/cim:id'),
        ('id', False, 'unicode', 'child::cim:requirement/cim:requirement/cim:initialCondition/cim:id'),
        ('id', False, 'unicode', 'child::cim:requirement/cim:requirement/cim:lateralBoundaryCondition/cim:id'),
        ('id', False, 'unicode', 'child::cim:requirement/cim:requirement/cim:outputRequirement/cim:id'),
        ('id', False, 'unicode', 'child::cim:requirement/cim:requirement/cim:spatioTemporalConstraint/cim:id'),
        ('name', False, 'unicode', 'child::cim:name'),
        ('name', False, 'unicode', 'child::cim:requirement/cim:requirement/cim:boundaryCondition/cim:name'),
        ('name', False, 'unicode', 'child::cim:requirement/cim:requirement/cim:initialCondition/cim:name'),
        ('name', False, 'unicode', 'child::cim:requirement/cim:requirement/cim:lateralBoundaryCondition/cim:name'),
        ('name', False, 'unicode', 'child::cim:requirement/cim:requirement/cim:outputRequirement/cim:name'),
        ('name', False, 'unicode', 'child::cim:requirement/cim:requirement/cim:spatioTemporalConstraint/cim:name'),
        ('relationship', False, 'unicode', 'self::cim:requirementOption/@optionRelationship'),
    ]

    return set_attributes(typeset.activity.NumericalRequirementOption(), xml, nsmap, decodings)


def decode_output_requirement(xml, nsmap):
    """Decodes an instance of the following type: output requirement.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.OutputRequirement

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('id', False, 'unicode', 'child::cim:id'),
        ('name', False, 'unicode', 'child::cim:name'),
        ('options', True, decode_numerical_requirement_option, 'child::cim:requirementOption'),
        ('source', False, decode_component_property, 'child::cim:source/cim:source/cim:componentProperty'),
        ('source', False, decode_data_content, 'child::cim:source/cim:source/cim:dataContent'),
        ('source', False, decode_data_object, 'child::cim:source/cim:source/cim:dataObject'),
        ('source', False, decode_doc_reference, 'child::cim:source/cim:reference'),
        ('source', False, decode_model_component, 'child::cim:source/cim:source/cim:softwareComponent'),
        ('source', False, decode_processor_component, 'child::cim:source/cim:source/cim:softwareComponent'),
        ('source', False, decode_statistical_model_component, 'child::cim:source/cim:source/cim:softwareComponent'),
    ]

    return set_attributes(typeset.activity.OutputRequirement(), xml, nsmap, decodings)


def decode_physical_modification(xml, nsmap):
    """Decodes an instance of the following type: physical modification.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.PhysicalModification

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('frequency', False, 'unicode', 'child::cim:frequency'),
        ('is_conformant', False, 'bool', '@conformant'),
        ('requirements', True, decode_boundary_condition, 'child::cim:requirement/cim:requirement/cim:boundaryCondition'),
        ('requirements', True, decode_doc_reference, 'child::cim:requirement/cim:reference'),
        ('requirements', True, decode_initial_condition, 'child::cim:requirement/cim:requirement/cim:initialCondition'),
        ('requirements', True, decode_lateral_boundary_condition, 'child::cim:requirement/cim:requirement/cim:lateralBoundaryCondition'),
        ('requirements', True, decode_output_requirement, 'child::cim:requirement/cim:requirement/cim:outputRequirement'),
        ('requirements', True, decode_spatio_temporal_constraint, 'child::cim:requirement/cim:requirement/cim:spatioTemporalConstraint'),
        ('sources', True, decode_component_property, 'child::cim:source/cim:source/cim:componentProperty'),
        ('sources', True, decode_data_content, 'child::cim:source/cim:source/cim:dataContent'),
        ('sources', True, decode_data_object, 'child::cim:source/cim:source/cim:dataObject'),
        ('sources', True, decode_doc_reference, 'child::cim:source/cim:reference'),
        ('sources', True, decode_model_component, 'child::cim:source/cim:source/cim:softwareComponent'),
        ('sources', True, decode_processor_component, 'child::cim:source/cim:source/cim:softwareComponent'),
        ('sources', True, decode_statistical_model_component, 'child::cim:source/cim:source/cim:softwareComponent'),
        ('type', False, 'unicode', '@type'),
    ]

    return set_attributes(typeset.activity.PhysicalModification(), xml, nsmap, decodings)


def decode_simulation(xml, nsmap):
    """Decodes an instance of the following type: simulation.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.Simulation

    """
    decodings = [
        ('authors', False, 'unicode', 'child::cim:authorsList/cim:list'),
        ('calendar', False, decode_daily_360, 'child::cim:calendar/cim:daily-360'),
        ('calendar', False, decode_perpetual_period, 'child::cim:calendar/cim:perpetualPeriod'),
        ('calendar', False, decode_real_calendar, 'child::cim:calendar/cim:realCalendar'),
        ('conformances', True, decode_conformance, 'child::cim:conformance/cim:conformance'),
        ('conformances', True, decode_physical_modification, 'child::cim:conformance/cim:physicalModification'),
        ('control_simulation', False, decode_doc_reference, 'child::cim:controlSimulation/cim:reference'),
        ('control_simulation', False, decode_simulation, 'child::cim:controlSimulation/cim:controlSimulation'),
        ('deployments', True, decode_deployment, 'child::cim:deployment'),
        ('description', False, 'unicode', 'child::cim:description'),
        ('funding_sources', True, 'unicode', 'child::cim:fundingSource'),
        ('inputs', True, decode_coupling, 'child::cim:input'),
        ('long_name', False, 'unicode', 'child::cim:longName'),
        ('projects', True, 'unicode', 'child::cim:project/@value'),
        ('rationales', True, 'unicode', 'child::cim:rationale'),
        ('responsible_parties', True, decode_responsible_party, 'child::cim:responsibleParty'),
        ('short_name', False, 'unicode', 'child::cim:shortName'),
        ('simulation_id', False, 'unicode', 'child::cim:simulationID'),
        ('spinup_date_range', False, decode_closed_date_range, 'child::cim:dateRange/cim:closedDateRange'),
        ('supports', True, decode_doc_reference, 'child::cim:supports/cim:reference'),
        ('supports', True, decode_experiment, 'child::cim:supports/cim:experiment'),
    ]

    return set_attributes(typeset.activity.Simulation(), xml, nsmap, decodings)


def decode_simulation_composite(xml, nsmap):
    """Decodes an instance of the following type: simulation composite.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.SimulationComposite

    """
    decodings = [
        ('authors', False, 'unicode', 'child::cim:authorsList/cim:list'),
        ('calendar', False, decode_daily_360, 'child::cim:calendar/cim:daily-360'),
        ('calendar', False, decode_perpetual_period, 'child::cim:calendar/cim:perpetualPeriod'),
        ('calendar', False, decode_real_calendar, 'child::cim:calendar/cim:realCalendar'),
        ('child', True, decode_simulation_composite, 'child::cim:child'),
        ('child', True, decode_simulation_run, 'child::cim:child'),
        ('conformances', True, decode_conformance, 'child::cim:conformance/cim:conformance'),
        ('conformances', True, decode_physical_modification, 'child::cim:conformance/cim:physicalModification'),
        ('control_simulation', False, decode_doc_reference, 'child::cim:controlSimulation/cim:reference'),
        ('control_simulation', False, decode_simulation, 'child::cim:controlSimulation/cim:controlSimulation'),
        ('date_range', False, decode_closed_date_range, 'child::cim:dateRange/cim:closedDateRange'),
        ('date_range', False, decode_open_date_range, 'child::cim:dateRange/cim:openDateRange'),
        ('deployments', True, decode_deployment, 'child::cim:deployment'),
        ('description', False, 'unicode', 'child::cim:description'),
        ('funding_sources', True, 'unicode', 'child::cim:fundingSource'),
        ('inputs', True, decode_coupling, 'child::cim:input'),
        ('long_name', False, 'unicode', 'child::cim:longName'),
        ('meta', False, decode_doc_meta_info, 'self::cim:simulationRun'),
        ('projects', True, 'unicode', 'child::cim:project/@value'),
        ('rank', False, 'int', 'child::cim:rank'),
        ('rationales', True, 'unicode', 'child::cim:rationale'),
        ('responsible_parties', True, decode_responsible_party, 'child::cim:responsibleParty'),
        ('short_name', False, 'unicode', 'child::cim:shortName'),
        ('simulation_id', False, 'unicode', 'child::cim:simulationID'),
        ('spinup_date_range', False, decode_closed_date_range, 'child::cim:dateRange/cim:closedDateRange'),
        ('supports', True, decode_doc_reference, 'child::cim:supports/cim:reference'),
        ('supports', True, decode_experiment, 'child::cim:supports/cim:experiment'),
    ]

    return set_attributes(typeset.activity.SimulationComposite(), xml, nsmap, decodings)


def decode_simulation_relationship(xml, nsmap):
    """Decodes an instance of the following type: simulation relationship.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.SimulationRelationship

    """
    decodings = [
    ]

    return set_attributes(typeset.activity.SimulationRelationship(), xml, nsmap, decodings)


def decode_simulation_relationship_target(xml, nsmap):
    """Decodes an instance of the following type: simulation relationship target.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.SimulationRelationshipTarget

    """
    decodings = [
    ]

    return set_attributes(typeset.activity.SimulationRelationshipTarget(), xml, nsmap, decodings)


def decode_simulation_run(xml, nsmap):
    """Decodes an instance of the following type: simulation run.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.SimulationRun

    """
    decodings = [
        ('authors', False, 'unicode', 'child::cim:authorsList/cim:list'),
        ('calendar', False, decode_daily_360, 'child::cim:calendar/cim:daily-360'),
        ('calendar', False, decode_perpetual_period, 'child::cim:calendar/cim:perpetualPeriod'),
        ('calendar', False, decode_real_calendar, 'child::cim:calendar/cim:realCalendar'),
        ('conformances', True, decode_conformance, 'child::cim:conformance/cim:conformance'),
        ('conformances', True, decode_physical_modification, 'child::cim:conformance/cim:physicalModification'),
        ('control_simulation', False, decode_doc_reference, 'child::cim:controlSimulation/cim:reference'),
        ('control_simulation', False, decode_simulation, 'child::cim:controlSimulation/cim:controlSimulation'),
        ('date_range', False, decode_closed_date_range, 'child::cim:dateRange/cim:closedDateRange'),
        ('date_range', False, decode_open_date_range, 'child::cim:dateRange/cim:openDateRange'),
        ('deployments', True, decode_deployment, 'child::cim:deployment'),
        ('description', False, 'unicode', 'child::cim:description'),
        ('funding_sources', True, 'unicode', 'child::cim:fundingSource'),
        ('inputs', True, decode_coupling, 'child::cim:input'),
        ('long_name', False, 'unicode', 'child::cim:longName'),
        ('meta', False, decode_doc_meta_info, 'self::cim:simulationRun'),
        ('model', False, decode_doc_reference, 'child::cim:model/cim:reference'),
        ('model', False, decode_model_component, 'child::cim:model/cim:modelComponent'),
        ('projects', True, 'unicode', 'child::cim:project/@value'),
        ('rationales', True, 'unicode', 'child::cim:rationale'),
        ('responsible_parties', True, decode_responsible_party, 'child::cim:responsibleParty'),
        ('short_name', False, 'unicode', 'child::cim:shortName'),
        ('simulation_id', False, 'unicode', 'child::cim:simulationID'),
        ('spinup_date_range', False, decode_closed_date_range, 'child::cim:dateRange/cim:closedDateRange'),
        ('supports', True, decode_doc_reference, 'child::cim:supports/cim:reference'),
        ('supports', True, decode_experiment, 'child::cim:supports/cim:experiment'),
    ]

    return set_attributes(typeset.activity.SimulationRun(), xml, nsmap, decodings)


def decode_spatio_temporal_constraint(xml, nsmap):
    """Decodes an instance of the following type: spatio temporal constraint.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.activity.SpatioTemporalConstraint

    """
    decodings = [
        ('date_range', False, decode_closed_date_range, 'child::cim:requiredDuration/cim:closedDateRange'),
        ('date_range', False, decode_open_date_range, 'child::cim:requiredDuration/cim:openDateRange'),
        ('description', False, 'unicode', 'child::cim:description'),
        ('id', False, 'unicode', 'child::cim:id'),
        ('name', False, 'unicode', 'child::cim:name'),
        ('options', True, decode_numerical_requirement_option, 'child::cim:requirementOption'),
        ('source', False, decode_component_property, 'child::cim:source/cim:source/cim:componentProperty'),
        ('source', False, decode_data_content, 'child::cim:source/cim:source/cim:dataContent'),
        ('source', False, decode_data_object, 'child::cim:source/cim:source/cim:dataObject'),
        ('source', False, decode_doc_reference, 'child::cim:source/cim:reference'),
        ('source', False, decode_model_component, 'child::cim:source/cim:source/cim:softwareComponent'),
        ('source', False, decode_processor_component, 'child::cim:source/cim:source/cim:softwareComponent'),
        ('source', False, decode_statistical_model_component, 'child::cim:source/cim:source/cim:softwareComponent'),
        ('spatial_resolution', False, 'unicode', 'child::cim:spatialResolution'),
    ]

    return set_attributes(typeset.activity.SpatioTemporalConstraint(), xml, nsmap, decodings)





# -*- coding: utf-8 -*-

"""
.. module:: model_topic.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Model topic notebook data wrapper.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import collections
import json
import os

from pyesdoc.ipython import constants
from pyesdoc.mp.specializations import get_model_specialization
from pyesdoc.mp.specializations import get_property_specialization



class NotebookOutput(object):
    """Model topic ipython data wrapper.

    """
    def __init__(self, mip_era, institute, source_id, topic):
        """Instance initialiser.

        """
        self.authors = []
        self.contributors = []
        self.content = dict()
        self.institute = unicode(institute).strip().lower()
        self.mip_era = unicode(mip_era).strip().lower()
        self.publication_status = 0
        self.seeding_source = None
        self.source_id = unicode(source_id).strip().lower()
        self.specialization = get_model_specialization(mip_era, topic)
        self.topic = unicode(topic).strip().lower()
        self._prop = None
        self._prop_specialization = None

        # Initialise output directory.
        if os.getenv('ESDOC_CMIP6_NOTEBOOK_HOME') is not None:
            output_dir = os.getenv('ESDOC_CMIP6_NOTEBOOK_HOME')
            output_dir = os.path.join(output_dir, self.institute)
        else:
            output_dir = os.path.expanduser('~')
            output_dir = os.path.join(output_dir, 'notebooks')
        output_dir = os.path.join(output_dir, self.mip_era)
        output_dir = os.path.join(output_dir, 'models')
        output_dir = os.path.join(output_dir, self.source_id)
        if not os.path.isdir(output_dir):
            os.makedirs(output_dir)

        # Initialise state from previously saved output file.
        self.fpath = os.path.join(output_dir, '.{}.json'.format(self.topic))
        if os.path.exists(self.fpath):
            with open(self.fpath, 'r') as fstream:
                self._from_dict(json.loads(fstream.read()))


    def save(self):
        """Persists state to file system.

        """
        obj = self._to_dict()
        with open(self.fpath, 'w') as fstream:
            fstream.write(json.dumps(obj, indent=4))


    def _from_dict(self, obj):
        """Initialises internal state from a dictionary.

        """
        self.publication_status = obj.get('publicationState', 0)
        self.mip_era = obj['mipEra']
        self.institute = obj['institute']
        self.seeding_source = obj.get('seedingSource')
        self.source_id = obj['sourceID']
        self.topic = obj['topic']
        self.authors = [(i['name'], i['email']) for i in obj['authors']]
        self.contributors = [(i['name'], i['email']) for i in obj['contributors']]
        self.content = obj['content']


    def _to_dict(self):
        """Returns a dictionary representation of internal state.

        """
        obj = collections.OrderedDict()
        obj['publicationState'] = self.publication_status
        obj['mipEra'] = self.mip_era
        obj['institute'] = self.institute
        obj['seedingSource'] = self.seeding_source
        obj['sourceID'] = self.source_id
        obj['topic'] = self.topic
        obj['authors'] = [{'name': i[0], 'email': i[1]} for i in self.authors]
        obj['contributors'] = [{'name': i[0], 'email': i[1]} for i in self.contributors]
        obj['content'] = collections.OrderedDict()
        for specialization_id in sorted(self.content.keys()):
            specialization_obj = self.content[specialization_id]
            if specialization_obj['qcStatus'] or specialization_obj['values']:
                obj['content'][specialization_id] = self.content[specialization_id]

        return obj


    def set_author(self, name, email):
        """Adds an author to managed collection.

        """
        # Format inputs.
        if name is not None:
            name = unicode(name).strip()
        if email is not None:
            email = unicode(email).strip()

        # Validate inputs.
        if name is None or len(name) == 0:
            raise ValueError('Invalid contributor name')
        if email is None or len(email) == 0:
            raise ValueError('Invalid contributor email')
        # TODO: validate with reg-ex.

        # Reject duplicates.
        for i, j in self.authors:
            if name.lower() == i.lower() and email.lower() == j.lower():
                return

        # Update state.
        self.authors.append((name, email))
        self.save()


    def set_contributor(self, name, email):
        """Adds a contributor to managed collection.

        """
        # Format inputs.
        if name is not None:
            name = unicode(name).strip()
        if email is not None:
            email = unicode(email).strip()

        # Validate inputs.
        if name is None or len(name) == 0:
            raise ValueError('Invalid contributor name')
        if email is None or len(email) == 0:
            raise ValueError('Invalid contributor email')
        # TODO: validate with reg-ex.

        # Reject duplicates.
        for i, j in self.contributors:
            if name.lower() == i.lower() and email.lower() == j.lower():
                return

        # Update state.
        self.contributors.append((name, email))
        self.save()


    def set_id(self, prop_id):
        """Sets id of specialized property being edited.

        """
        self.content[prop_id] = {
            'values': [],
            'qcStatus': 0
        }
        self._prop = self.content[prop_id]
        self._prop_specialization = get_property_specialization(prop_id)


    def set_qc_status(self, qc_status):
        """Sets qc-status of specialized property being edited.

        """
        # Validate input.
        if qc_status not in constants.QC_STATES:
            raise ValueError('Invalid QC status')

        # Update state.
        self._prop['qcStatus'] = qc_status
        self.save()


    def set_publication_status(self, publication_status):
        """Sets publication status of document being edited.

        """
        # Validate input.
        if publication_status not in constants.PUBLICATION_STATES:
            raise ValueError('Invalid publication status')

        # Update state.
        self.publication_status = publication_status
        self.save()


    def set_value(self, val):
        """Sets a scalar value.

        """
        # Validate input:
        # ... error if trying to add > 1 value to a property with singular cardinality.
        if not self._prop_specialization.is_collection and \
           len(self._prop['values']) >= 1:
            raise ValueError('Invalid property: only one value can be added')

        # ... error if adding a duplicate value.
        if val in self._prop['values']:
            raise ValueError('Invalid property: cannot add duplicate values')

        # ... error if specialization complains.
        self._prop_specialization.validate_value(val)

        # Update state.
        self._prop['values'].append(val)
        self.save()


    def get_values(self, specialization_id):
        """Returns a set of values.

        """
        return self.content.get(specialization_id, dict()).get('values', [])


    def get_qc_status(self, specialization_id):
        """Returns a property qc status.

        """
        return self.content.get(specialization_id, dict()).get('qcStatus', 0)




# -*- coding: utf-8 -*-

"""
.. module:: test_general.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyesdoc general tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
# Module imports.
import pyesdoc
from pyesdoc.ontologies import cim
import test_utils as tu
import test_types as tt



# Test ontology constants.
_CIM = 'cim'
_CIM_V1 = '1'
_CIM_PACKAGE = 'software'
_CIM_TYPE = 'modelComponent'

# Test constants.
_INSTITUTE = 'TEST'
_PROJECT = 'TEST'


def _create_doc(doc_type=cim.v1.ModelComponent):
    """Creates a test document."""
    return pyesdoc.create(doc_type, _INSTITUTE, _PROJECT)


def _assert_doc(doc, typeof=None):
    """Perform standard test document assertions."""
    tu.assert_object(doc, typeof)
    if hasattr(doc, 'meta'):
        tu.assert_str(doc.meta.institute, _INSTITUTE.lower())
        tu.assert_str(doc.meta.project, _PROJECT.lower())
        tu.assert_str(doc.meta.source, pyesdoc.DEFAULT_SOURCE)
        if typeof is not None:
            tu.assert_str(doc.meta.type, typeof.type_key)


def _test_create_01():
    """Test creating documents - 1."""
    doc = _create_doc()

    _assert_doc(doc, cim.v1.ModelComponent)


def _test_create_02():
    """Test creating documents - 3."""
    for doc_type in pyesdoc.get_types():
        doc = pyesdoc.create(doc_type, institute=_INSTITUTE, project=_PROJECT)
        _assert_doc(doc, doc_type)


def test():
    """Runs set of factory unit tests."""
    for func in (
        _test_create_01,
        _test_create_02,
        ):
        tu.init(func, func.__doc__[5:])
        yield func


"""
A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.
"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Ruth Petrie'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Ruth Petrie, Bryan Lawrence'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Sea Ice Dynamics'

# --------------------------------------------------------------------
# DETAILS : General process details.
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'Dynamic processes in sea ice.',
    'properties': [
        ('horizontal_transport', 'ENUM:transport_methods', '1.1',
            "What is the method of horizontal advection of sea ice?"),
        ('transport_in_thickness_space', 'ENUM:transport_methods', '1.1',
            "What is the method of sea ice transport in thickness space (i.e. in thickness categories)?"),
        ('ice_strength_formulation', 'ENUM:ice_strength', '1.1',
             "Which method of sea ice strength formulation is used?"),
        ('redistribution', 'ENUM:redistribution_types', '1.N',
             "Which processes can redistribute sea ice?"),
        ('rheology', 'ENUM:rheology_types', '1.1',
             "Rheology, what is the ice deformation formulation?"),
        ]
    }

# --------------------------------------------------------------------
# PROCESS: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['transport_methods'] = {
    'description': 'Transport Methods',
    'is_open': True,
    'members': [
        ('Incremental Re-mapping', '(including Semi-Lagrangian)'),
        ('Prather', None),
        ('Eulerian', None),
    ]
}

ENUMERATIONS['redistribution_types'] = {
    'description': 'Sea Ice Redistribution types',
    'is_open': True,
    'members': [
        ('Rafting', None),
        ('Ridging', None),
    ]
}

ENUMERATIONS['ice_strength'] = {
    'description': 'Ice strength formulation methods',
    'is_open': True,
    'members': [
        ('Hibler 1979', None),
        ('Rothrock 1975', None),
    ]
}

ENUMERATIONS['rheology_types'] = {
    'description': 'Sea ice rheology types',
    'is_open': True,
    'members': [
        ('Free-drift', None),
        ('Mohr-Coloumb', None),
        ('Visco-plastic', 'VP'),
        ('Elastic-visco-plastic', 'EVP'),
        ('Elastic-aniostropic-plastic', None,),
        ('Granular', None),
    ]
}





# -*- coding: utf-8 -*-

"""
.. module:: software_classes.py
   :synopsis: Set of CIM v2 ontology type definitions.

"""


def component_base():
    """Base class for software component properties, whether a top level model,
    or a specific piece of code known as a component. In software terms, a
    component is a discrete set of code that takes input data and generates output data.
    Components may or may not have scientific descriptions.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': True,
        'pstr': ('{}', ('name',)), 
        'properties': [
            ('citations', 'shared.citation', '0.N',
                "Set of pertinent citations."),
            ('canonical_id', 'str', '0.1',
                "Vocabulary identifier, where this framework/component description was constructed via a controlled vocabulary."),
            ('description', 'str', '0.1',
                "Textural description of component."),
            ('long_name', 'str', '0.1',
                "Long name for component."),
            ('name', 'str', '1.1',
                "Short name of component."),
            ('responsible_parties', 'shared.responsibility', '0.N',
                "People or organisations responsible for providing this information."),

            ('development_history', 'software.development_path', '0.1',
                "History of the development of this component."),
            ('release_date', 'datetime', '0.1',
                "The date of publication of the component code (as opposed to the date of publication of the metadata document, or the date of deployment of the model)."),
            ('repository', 'shared.online_resource', '0.1',
                "Location of code for this component."),
            ('version', 'str', '0.1',
                "Version identifier."),
        ]
    }



def implementation():
    """Implementation information for a software framework/component, whether a top level model,
    or a specific piece of code known as a 'component'. In software terms, a
    software framework/component is a discrete set of code that takes input data and generates output data.
    Software frameworks/components may or may not have scientific descriptions.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'pstr': ('{}', ('name',)),
        'properties': [
            ('citations', 'shared.citation', '0.N',
                "Set of pertinent citations."),
            ('canonical_id', 'str', '0.1',
                "Vocabulary identifier, where this framework/component description was constructed via a controlled vocabulary."),
            ('description', 'str', '0.1',
                "Textural description of framework/component."),
            ('development_history', 'software.development_path', '0.1',
                "History of the development of this framework/component."),
            ('long_name', 'str', '0.1',
                "Long name for framework/component."),
            ('name', 'str', '1.1',
                "Short name of framework/component."),
            ('release_date', 'datetime', '0.1',
                "The date of publication of the framework/component code."),
            ('repository', 'shared.online_resource', '0.1',
                "Location of code for this framework/component."),
            ('version', 'str', '0.1',
                "Version identifier.")
        ]
    }


def composition():
    """Describes how component variables are coupled together either to/from other
    SoftwareComponents or external data files. The variables specified by a component's
    composition must be owned by that component, or a  child of that component;
    child components cannot couple together parent variables.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('couplings', 'str', '0.N',
                "#FIXME."),
            ('description', 'str', '0.1',
                "#FIXME.")
        ]
    }


def development_path():
    """Describes the software development path for this model/component.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('consortium_name', 'str', '0.1',
                "If model/component is developed as part of a consortium, provide consortium name."),
            ('creators', 'shared.responsibility', '0.N',
                "Those responsible for creating this component."),
            ('was_developed_in_house', 'bool', '1.1',
                "Model or component was mostly developed in house."),
            ('funding_sources', 'shared.responsibility', '0.N',
                "The entities that funded this software component."),
            ('previous_version', 'str', '0.1',
                "Name of a previous version.")
        ]
    }


def entry_point():
    """Describes a function or subroutine of a SoftwareComponent.
    BFG will use these EntryPoints to define a schedule of subroutine calls for a coupled model.
    Currently, a very basic schedule can be approximated by using the 'proceeds' and 'follows' attributes,
    however a more complete system is required for full BFG compatibility.
    Every EntryPoint can have a set of arguments associated with it.
    These reference (previously defined) variables.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('name', 'str', '0.1',
                "#FIXME.")
        ]
    }


def gridspec():
    """Fully defines the computational grid used.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('description', 'str', '1.1',
                "Textural description.")
        ]
    }


def software_component():
    """An embedded piece of software that does not normally function as a standalone model (although
    it may be used standalone in a test harness).

    """
    return {
        'type': 'class',
        'base': 'software.component_base',
        'is_abstract': False,
        'properties': [
            ('composition', 'software.composition', '0.1',
                "#FIXME."),
            ('connection_points', 'software.variable', '0.N',
                "The set of data entities which are available for I/O and/or coupling."),
            ('coupling_framework', 'software.coupling_framework', '0.1',
                "The coupling framework that this entire component conforms to."),
            ('dependencies', 'software.entry_point', '0.N',
                "#FIXME."),
            ('grid', 'software.gridspec', '0.1',
                "A reference to the grid that is used by this component."),
            ('language', 'software.programming_language', '0.1',
                "Language the component is written in."),
            ('license', 'str', '0.1',
                "The license held by this piece of software."),
            ('sub_components', 'software.software_component', '0.N',
                "Internal software sub-components of this component."),
            ('depends_on', 'software.software_component', '0.N',        # added dch/ssw 1016-0804
                "The software components whose outputs are inputs to this software component."),
            # Would like to think about making this a stand-alone document
        ]
    }


def variable():
    """An instance of a model software variable which may be prognostic or diagnostic, and which is
    available as a connection to other software components. Note that these variables may only exist
    within the software workflow as interim quantities or coupling endpoints. Input and output
    variables will be a subset of these software variables.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('description', 'str', '0.1',
                "Description of how the variable is being used in the s/w."),
            ('name', 'str', '1.1',
                "Short name for the variable."),
            ('is_prognostic', 'bool', '1.1',
                "Whether or not prognostic or diagnostic.")
        ]
    }



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._utils.logger.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Package logging utility functions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import arrow

from pyesdoc.utils.compat import str



# Set of logging levels.
LOG_LEVEL_DEBUG = 'DUBUG'
LOG_LEVEL_INFO = 'INFO'
LOG_LEVEL_WARNING = 'WARNING'
LOG_LEVEL_ERROR = 'ERROR'
LOG_LEVEL_CRITICAL = 'CRITICAL'
LOG_LEVEL_FATAL = 'FATAL'

# Defaults.
_DEFAULT_APP = 'PYESDOC'
_DEFAULT_INSTITUTE = 'ESDOC'

# Text to display when passed a null message.
_NULL_MSG = '-------------------------------------------------------------------------------'


def _get_formatted_message(msg, level, app):
    """Returns a message formatted for logging.

    """
    if msg is None:
        return _NULL_MSG
    else:
        return '{} [{}] :: ESDOC-{} :: {}'.format(
            str(arrow.get())[0:-6],
            level,
            app,
            str(msg).strip()
            )


def log(msg=None, level=LOG_LEVEL_INFO, app=_DEFAULT_APP):
    """Outputs a message to log.

    :param str msg: Message to be written to log.
    :param str level: Message level (e.g. INFO).
    :param str app: Application emitting log message (e.g. libIGCM).

    """
    # TODO use structlog/logstash.
    print(_get_formatted_message(msg, level, app))


def log_error(err, app=_DEFAULT_APP):
    """Logs a runtime error.

    :param str err: Error to be written to log.
    :param str app: Application emitting log message (e.g. libIGCM).

    """
    msg = '!! RUNTIME ERROR !! :: '
    if issubclass(BaseException, err.__class__):
        msg += '{} :: {}'.format(err.__class__, err)
    else:
        msg += '{}'.format(err)
    log(msg, LOG_LEVEL_ERROR, app)


def log_warning(err, app=_DEFAULT_APP):
    """Logs a runtime warning.

    :param str err: Error to be written to log.
    :param str app: Application emitting log message (e.g. libIGCM).

    """
    if issubclass(BaseException, err.__class__):
        msg = '{} :: {}'.format(err.__class__, err)
    else:
        msg = '{}'.format(err)
    log(msg, LOG_LEVEL_WARNING, app)



"""A realm key-properties sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Land ice key properties'

# --------------------------------------------------------------------
# KEY PROPERTIES: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'General key properties',
    'properties': [
        ('model_name', 'str', '1.1',
             'Name of land surface model code'),
        ('ice_albedo', 'ENUM:ice_albedo_methods', '1.N',
            'Specify how ice albedo is modelled',),
        ('atmospheric_coupling_variables', 'str', '1.1',
             'Which variables are passed between the atmosphere and ice (e.g. orography, ice mass)'), 
        ('oceanic_coupling_variables', 'str', '1.1',
             'Which variables are passed between the ocean and ice'), 
        ('prognostic_variables', 'ENUM:prognostic_variable_types', '1.N',
             'Which variables are prognostically calculated in the ice model'),
        ]
    }

# --------------------------------------------------------------------
# KEY PROPERTIES: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['ice_albedo_methods'] = {
    'description': 'Specify how ice albedo is modelled',
    'is_open': True,
    'members': [
        ('prescribed', None),
        ('function of ice age', None),
        ('function of ice density', None),
    ]
}

ENUMERATIONS['prognostic_variable_types'] = {
    'description': 'Which variables are prognostically calculated in the ice model',
    'is_open': True,
    'members': [
        ('ice velocity', None),
        ('ice thickness', None),
        ('ice temperature', None),
    ]
}





# -*- coding: utf-8 -*-

"""
.. module:: test_publishing.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyesdoc document publishing tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import uuid

import arrow
import pyesdoc

import test_utils as tu
import test_types as tt


# Testing mode.
_MODE = 'dev'



def _create_doc(mod):
    """Creates a document for testing."""
    doc = tu.get_doc(mod)
    mod.DOC_ID = doc.meta.id = uuid.uuid4()
    mod.DOC_VERSION = doc.meta.version = 0
    mod.DOC_DATE = doc.meta.create_date = doc.meta.update_date = arrow.now().datetime

    return doc


def _set_api_url():
    """Assigns the base API url."""
    if _MODE == 'dev':
        pyesdoc.set_option("api_url", "http://127.0.0.1:5000")
    elif _MODE == 'test':
        pyesdoc.set_option("api_url", "https://test-api.es-doc.org")
    elif _MODE == 'prod':
        pyesdoc.set_option("api_url", "https://api.es-doc.org")


def _test(mod):
    """Performs a publishing workflow test."""
    # Create document.
    doc = _create_doc(mod)

    # for err in pyesdoc.validate(doc):
    #     print err

    # Publish.
    pyesdoc.publish(doc)
    mod.DOC_VERSION += 1
    mod.assert_doc(doc)

    # Retrieve.
    doc = pyesdoc.retrieve(mod.DOC_ID, mod.DOC_VERSION)
    mod.assert_doc(doc)

    # Update.
    mod.update_doc(doc)
    mod.assert_doc_updates(doc)

    # Republish.
    pyesdoc.publish(doc)
    mod.DOC_VERSION += 1

    # Retrieve.
    doc = pyesdoc.retrieve(mod.DOC_ID, mod.DOC_VERSION)
    mod.assert_doc(doc, True)
    mod.assert_doc_updates(doc)

    # Republish.
    pyesdoc.publish(doc)
    mod.DOC_VERSION += 1

    # Unpublish.
    pyesdoc.unpublish(mod.DOC_ID, mod.DOC_VERSION)
    tu.assert_none(pyesdoc.retrieve(mod.DOC_ID, mod.DOC_VERSION))
    mod.DOC_VERSION -= 1

    # Retrieve specific.
    doc = pyesdoc.retrieve(mod.DOC_ID, mod.DOC_VERSION)
    mod.assert_doc(doc, True)
    mod.assert_doc_updates(doc)

    # Retrieve latest.
    doc = pyesdoc.retrieve(mod.DOC_ID, pyesdoc.DOC_VERSION_LATEST)
    mod.assert_doc(doc, True)
    mod.assert_doc_updates(doc)
    return

    # Unpublish.
    pyesdoc.unpublish(mod.DOC_ID, mod.DOC_VERSION)
    tu.assert_none(pyesdoc.retrieve(mod.DOC_ID, mod.DOC_VERSION))
    mod.DOC_VERSION -= 1

    # Retrieve specific.
    doc = pyesdoc.retrieve(mod.DOC_ID, mod.DOC_VERSION)
    mod.assert_doc(doc)

    # Retrieve latest.
    doc = pyesdoc.retrieve(mod.DOC_ID, pyesdoc.DOC_VERSION_LATEST)
    mod.assert_doc(doc)

    # Unpublish (will delete document altogether).
    pyesdoc.unpublish(mod.DOC_ID, mod.DOC_VERSION)
    tu.assert_none(pyesdoc.retrieve(mod.DOC_ID, mod.DOC_VERSION))
    mod.DOC_VERSION -= 1
    tu.assert_none(pyesdoc.retrieve(mod.DOC_ID, mod.DOC_VERSION))


def test():
    """Performs publishing workflow tests over the set of test documents."""
    _set_api_url()
    for mod in tt.MODULES:
        tu.init(_test, 'publishing', mod)
        yield _test, mod



# -*- coding: utf-8 -*-

"""
.. module:: data_classes.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Set of CIM v1 data package class definitions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""



def data_content():
    """The contents of the data object; like ISO: MD_ContentInformation.

    """
    return {
        'type' : 'class',
        'base' : 'shared.data_source',
        'is_abstract' : False,
        'properties' : [
            ('aggregation', 'str', '0.1'),
            ('frequency', 'str', '0.1'),
            ('topic', 'data.data_topic', '1.1'),
        ],
        'doc_strings': {
            'aggregation': 'Describes how the content has been aggregated together: sum, min, mean, max, ...',
            'frequency': 'Describes the frequency of the data content: daily, hourly, ...',
        },
        'decodings' : [
            ('aggregation', 'child::cim:aggregation'),
            ('frequency', 'child::cim:frequency/@value'),
            ('topic', 'child::cim:topic'),
        ]
    }


def data_distribution():
    """Describes how a DataObject is distributed.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('access', 'str', '0.1'),
            ('fee', 'str', '0.1'),
            ('format', 'str', '0.1'),
            ('responsible_party', 'shared.responsible_party', '0.1'),
        ],
        'decodings' : [
            ('access', '@distributionAccess'),
            ('format', 'child::cim:distributionFormat/@value'),
            ('responsible_party', 'child::cim:responsibleParty'),
        ]
    }


def data_extent():
    """A data object extent represents the geographical and temporal coverage associated with a data object.

    """

    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('temporal', 'data.data_extent_temporal', '1.1'),
            ('geographical', 'data.data_extent_geographical', '1.1'),
        ],
        'decodings' : [
            ('geographical', 'child::gmd:geographicElement'),
            ('temporal', 'child::gmd:temporalElement/gmd:EX_TemporalExtent/gmd:extent/gml:TimePeriod'),
        ]
    }


def data_extent_geographical():
    """A data object geographical extent represents the geographical coverage associated with a data object.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('east', 'float', '0.1'),
            ('south', 'float', '0.1'),
            ('west', 'float', '0.1'),
            ('north', 'float', '0.1'),
        ],
        'decodings' : [
            ('east', 'child::gmd:EX_GeographicBoundingBox/gmd:eastBoundLongitude/gco:Decimal'),
            ('south', 'child::gmd:EX_GeographicBoundingBox/gmd:southBoundLatitude/gco:Decimal'),
            ('west', 'child::gmd:EX_GeographicBoundingBox/gmd:westBoundLongitude/gco:Decimal'),
            ('north', 'child::gmd:EX_GeographicBoundingBox/gmd:northBoundLatitude/gco:Decimal'),
        ]
    }


def data_extent_temporal():
    """A data object temporal extent represents the temporal coverage associated with a data object.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('begin', 'date', '0.1'),
            ('end', 'date', '0.1'),
            ('time_interval', 'data.data_extent_time_interval', '0.1'),
        ],
        'decodings' : [
            ('begin', 'child::gml:beginPosition'),
            ('end', 'child::gml:endPosition'),
            ('time_interval', 'child::gml:timeInterval'),
        ]
    }


def data_extent_time_interval():
    """A data object temporal extent represents the temporal coverage associated with a data object.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('unit', 'str', '0.1'),
            ('factor', 'int', '0.1'),
            ('radix', 'int', '0.1'),
        ],
        'decodings' : [
            ('factor', '@factor'),
            ('radix', '@radix'),
            ('unit', '@unit'),
        ]
    }


def data_hierarchy_level():
    """The type of data object that is grouped together into a particular hierarchy.  Currently, this is made up of terms describing how the Met Office splits up archived data and how THREDDS categorises variables.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('name', 'data.data_hierarchy_type', '0.1'),
            ('is_open', 'bool', '0.1'),
            ('value', 'str', '0.1'),
        ],
        'doc_strings': {
            'name': 'What level in the data hierarchy (constructed by the self-referential parent/child aggregations) is this DataObject.',
            'value': 'What is the name of the specific HierarchyLevel this DataObject is being organised at (ie: if the HierarchyLevel is "run" then the name might be the runid).',
        },
        'decodings' : [
            ('is_open', 'child::cim:hierarchyLevelName/@open'),
            ('name', 'child::cim:hierarchyLevelName/@value'),
            ('value', 'child::cim:hierarchyLevelValue'),
        ]
    }


def data_object():
    """A DataObject describes a unit of data.  DataObjects can be grouped hierarchically.  The attributes hierarchyLevelName and hierarchyLevelValue describe how objects are grouped.

    """
    return {
        'type' : 'class',
        'base' : 'shared.data_source',
        'is_abstract' : False,
        'is_document': True,
        'properties' : [
            ('acronym', 'str', '0.1'),
            ('child_object', 'data.data_object', '0.N'),
            ('citations', 'shared.citation', '0.N'),
            ('content', 'data.data_content', '0.N'),
            ('data_status', 'data.data_status_type', '0.1'),
            ('description', 'str', '0.1'),
            ('distribution', 'data.data_distribution', '0.1'),
            ('extent', 'data.data_extent', '0.1'),
            ('geometry_model', 'str', '0.1'),
            ('hierarchy_level', 'data.data_hierarchy_level', '0.1'),
            ('keyword', 'str', '0.1'),
            ('parent_object', 'data.data_object', '0.1'),
            ('properties', 'data.data_property', '0.N'),
            ('purpose', 'str', '0.1'),
            ('restriction', 'data.data_restriction', '0.N'),
            ('source_simulation', 'str', '0.1'),
            ('storage', 'data.data_storage', '0.N'),
        ],
        'doc_strings': {
            'content': 'The content of a DataObject corresponds to a variable (in THREDDS, ...etc.)'
        },
        'decodings' : [
            ('meta', 'self::cim:dataObject'),
            ('acronym', 'child::cim:acronym'),
            ('citations', '//cim:citation[not(cim:citation)]'),
            ('content', 'child::cim:content'),
            ('data_status', 'self::cim:dataObject/@dataStatus'),
            ('description', 'child::cim:description'),
            ('distribution', 'child::cim:distribution'),
            ('extent', 'child::cim:extent'),
            ('hierarchy_level', 'self::cim:dataObject'),
            ('keyword', 'child::cim:keyword'),
            ('properties', 'child::cim:dataProperty/cim:dataProperty'),
            ('purpose', 'self::cim:dataObject/@purpose'),
            ('storage', 'child::cim:storage/cim:ipStorage', 'data.data_storage_ip'),
        ]
    }


def data_property():
    """A property of a DataObject. Currently this is intended to be used to record CF specific information (like packing, scaling, etc.) for OASIS4.

    """
    return {
        'type' : 'class',
        'base' : 'shared.property',
        'is_abstract' : False,
        'properties' : [
            ('description', 'str', '0.1'),
        ],
        'decodings' : [
            ('description', 'child::cim:description'),
        ]
    }


def data_restriction():
    """An access or use restriction on some element of the DataObject actual data.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('scope', 'str', '0.1'),
            ('restriction', 'str', '0.1'),
            ('license', 'shared.license', '0.1'),
        ],
        'doc_strings': {
            'scope': 'The thing (data or metadata, access or use) that this restriction is applied to.',
            'restriction': 'The thing (data or metadata, access or use) that this restriction is applied to.',
            'license': 'The thing (data or metadata, access or use) that this restriction is applied to.'
        },
        'decodings' : [

        ]
    }


def data_storage():
    """Describes the method that the DataObject is stored. An abstract class with specific child classes for each supported method.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : True,
        'properties' : [
            ('size', 'int', '0.1'),
            ('format', 'str', '0.1'),
            ('modification_date', 'datetime', '0.1'),
            ('location', 'str', '0.1'),
        ],
        'decodings' : [
        ]
    }


def data_storage_db():
    """Contains attributes to describe a DataObject stored as a database file.

    """
    return {
        'type' : 'class',
        'base' : 'data.data_storage',
        'is_abstract' : False,
        'properties' : [
            ('access_string', 'str', '0.1'),
            ('name', 'str', '0.1'),
            ('owner', 'str', '0.1'),
            ('table', 'str', '0.1'),
        ],
        'decodings' : [

        ]
    }


def data_storage_file():
    """Contains attributes to describe a DataObject stored as a single file.

    """
    return {
        'type' : 'class',
        'base' : 'data.data_storage',
        'is_abstract' : False,
        'properties' : [
            ('file_system', 'str', '0.1'),
            ('path', 'str', '0.1'),
            ('file_name', 'str', '0.1'),
        ],
        'decodings' : [

        ]
    }


def data_storage_ip():
    """Contains attributes to describe a DataObject stored as a database file.

    """
    return {
        'type' : 'class',
        'base' : 'data.data_storage',
        'is_abstract' : False,
        'properties' : [
            ('protocol', 'str', '0.1'),
            ('host', 'str', '0.1'),
            ('path', 'str', '0.1'),
            ('file_name', 'str', '0.1'),
        ],
        'decodings' : [
            ('file_name', 'child::cim:fileName'),
            ('format', 'child::cim:dataFormat/@value'),
        ]
    }


def data_topic():
    """Describes the content of a data object: the variable name, units, etc.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('name', 'str', '0.1'),
            ('description', 'str', '0.1'),
            ('unit', 'str', '0.1'),
        ],
        'decodings' : [
            ('description', 'child::cim:description'),
            ('name', 'child::cim:name'),
            ('unit', 'child::cim:unit/@value'),
        ]
    }



# -*- coding: utf-8 -*-

"""
.. module:: shared_classes_time.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Set of CIM v1 shared package time related class definitions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""



def calendar():
    """Describes a method of calculating a span of dates.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': True,
        'properties': [
            ('description', 'str', '0.1'),
            ('length', 'int', '0.1'),
            ('range', 'shared.date_range', '0.1'),
        ],
        'doc_strings': {
            'description': 'Describes the finer details of the calendar, in case they are not-obvious.  For example, if an experiment has changing conditions within it (ie: 1% CO2 increase until 2100, then hold fixed for the remaining period of the  experment)',
        },
        'decodings': [
            ('description', 'child::cim:description'),
            ('length', 'child::cim:length'),
            ('range', 'child::cim:range/cim:closedDateRange', 'shared.closed_date_range'),
            ('range', 'child::cim:range/cim:openDateRange', 'shared.open_date_range'),
        ]
    }


def closed_date_range():
    """A date range with specified start and end points.

    """
    return {
        'type': 'class',
        'base': 'shared.date_range',
        'is_abstract': False,
        'properties': [
            ('end', 'datetime', '0.1'),
            ('start', 'datetime', '1.1'),
        ],
        'doc_strings': {
            'end': 'End date is optional becuase the length of a ClosedDateRange can be calculated from the StartDate plus the Duration element.',
        },
        'decodings': [
            ('end', 'child::cim:endDate'),
            ('start', 'child::cim:startDate'),
        ]
    }


def daily_360():
    """Creates and returns instance of daily_360 class.

    """
    return {
        'type': 'class',
        'base': 'shared.calendar',
        'is_abstract': False
    }


def date_range():
    """Creates and returns instance of date_range class.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': True,
        'properties': [
            ('duration', 'str', '0.1'),
        ],
        'decodings': [
            ('duration', 'child::cim:duration'),
        ]
    }


def open_date_range():
    """A date range without a specified start and/or end point.

    """
    return {
        'type': 'class',
        'base': 'shared.date_range',
        'is_abstract': False,
        'properties': [
            ('end', 'datetime', '0.1'),
            ('start', 'datetime', '0.1'),
        ],
        'decodings': [
            ('end', 'child::cim:endDate'),
            ('start', 'child::cim:startDate'),
        ]
    }


def perpetual_period():
    """Creates and returns instance of perpetual_period class.

    """
    return {
        'type': 'class',
        'base': 'shared.calendar',
        'is_abstract': False
    }


def real_calendar():
    """Creates and returns instance of real_calendar class.

    """
    return {
        'type': 'class',
        'base': 'shared.calendar',
        'is_abstract': False
    }



# -*- coding: utf-8 -*-

"""
.. module:: test_type_cim_v1_misc_document_set.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Tests a cim.v1.DocumentSet instance.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import arrow

import pyesdoc.ontologies.cim as cim
import test_utils as tu



# Test type.
DOC_TYPE = cim.v1.DocumentSet

# Test document type.
DOC_TYPE_KEY = DOC_TYPE.type_key

# Test representation file.
DOC_FILE = 'cim.1.misc.DocumentSet.xml-metafor-cim-v1'

# Test document uid.
DOC_UID = "e4155a28-268e-11e1-9a0e-00163e9152a5"

# Test document version.
DOC_VERSION = 2

# Test document creation date.
DOC_DATE = arrow.get("2012-04-23 14:59:06.767729").datetime

# Test document project.
DOC_PROJECT = "CMIP5"

# Test document project.
DOC_INSTITUTE = "MOHC"

# Test document author.
DOC_AUTHOR = "Metafor Questionnaire"


def assert_doc(doc, is_update=False):
    """Asserts a document.

    :param object doc: Document being tested.
    :param bool is_update: Flag indicating whether document has been updated.

    """
    tu.assert_iter(doc.data, 18, cim.v1.DataObject)
    tu.assert_iter(doc.ensembles, 1, cim.v1.Ensemble)
    tu.assert_object(doc.experiment, cim.v1.NumericalExperiment)
    tu.assert_iter(doc.grids, 1, cim.v1.GridSpec)
    tu.assert_object(doc.model, cim.v1.ModelComponent)
    tu.assert_object(doc.platform, cim.v1.Platform)
    tu.assert_object(doc.simulation, cim.v1.SimulationRun)


def update_doc(doc):
    """Update a document prior to republishing.

    :param object doc: Document being republished.

    """
    pass


def assert_doc_updates(doc):
    """Asserts a document after being updated.

    :param object doc: Document being tested.

    """
    pass